/* 
 * Shapable.java 1.0 2001-01-21
 *
 *  2001 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.appearance;

import java.awt.Shape;
import java.awt.Stroke;
import java.awt.Paint;
import java.awt.Color;

/**
 *  Provides definitions for objects that can be displayed
 *  in the <code>Graphics2D</code> context. When an object
 *  implementing the <code>Shapable</code> inteface is displayed,
 *  the methods are used followingly:
 *        <blockquote><code> 
 *        Graphics2D g2;<br>
 *        Shapable s;<br>
 *        </code>...<code><br>
 *        if (s.isFilled()) {<br>
 *        &nbsp;&nbsp;&nbsp;&nbsp;g2.setPaint(s.paint());<br>
 *        &nbsp;&nbsp;&nbsp;&nbsp;g2.fill(s.shape());<br>
 *        }<br>
 *        if (s.isDrawn()) {<br>
 *        &nbsp;&nbsp;&nbsp;&nbsp;g2.setPaint(s.color());<br>
 *        &nbsp;&nbsp;&nbsp;&nbsp;g2.setStroke(s.stroke());<br>
 *        &nbsp;&nbsp;&nbsp;&nbsp;g2.draw(s.shape());<br>
 *        }<br>
 *        </code></blockquote>
 *
 *  For further information on the <code>Graphics2D</code> class,
 *  see the trail
 *  <a href="http://java.sun.com/docs/books/tutorial/2d/">
 *  "2D Graphics"</a> in <em>The Java Tutorial</em>.
 *
 *  @see java.awt.Graphics2D
 *  @version 1.0, 010121
 *  @author  Jouni Smed
 */

public interface Shapable
  {
  /**
   *  Returns a <code>Shape</code> object that defines the appearance
   *  of the object.
   *  @see java.awt.Shape
   */
  public Shape shape();

  /** 
   *  Returns a boolean value indicating whether the outline of the
   *  shape is drawn or not.
   *  @see #stroke()
   *  @see #color()
   */ 
  public boolean isDrawn();
  
  /**
   *  Returns a <code>Stroke</code> object that is used for drawing
   *  the outline. 
   *  @see #isDrawn()
   *  @see java.awt.Stroke
   */
  public Stroke stroke();

  /**
   *  Returns a <code>Color</code> object that is used for drawing
   *  the outline. 
   *  @see #isDrawn()
   *  @see java.awt.Color
   */
  public Color color();

  /**
   *  Returns a boolean value indicating whether the shape is filled
   *  or not.
   *  @see #paint()
   */
  public boolean isFilled();
  /**
   *  Returns a <code>Paint</code> object that is used for filling
   *  the shape. 
   *  @see #isFilled()
   *  @see java.awt.Paint
   */
  public Paint paint();
  }
