/* 
 * Player.java 1.0 2002-08-23
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
package fi.utu.cs.hockey;

import fi.utu.cs.physics.*;
import fi.utu.cs.appearance.*;
import java.awt.Shape;
import java.awt.Paint;
import java.awt.Color;
import java.awt.Stroke;
import java.awt.GradientPaint;
import java.awt.BasicStroke;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Arc2D;
import java.awt.geom.Area;
import java.awt.Polygon;

/**
 *  A player. 
 *
 *  @version 1.0, 020823
 *  @author  Jouni Smed
 */
public class Player extends OrientedBody implements MultiShapable
  {
  /** Radius of a player in meters. */
  public static final double RADIUS = 0.35;
  /** Reachability of a player in meters. */
	public static final double REACHABILITY = 0.5;
	/** Mass of the player in kilograms. */
  private static final double MASS = 75.0;
  
	/** The maximum power of a forward dash in a second. */
	private static final double MAX_DASH_POWER_PER_SECOND = 0.25;
	/** The maximum power of a backward dash in a second. */
	private static final double MAX_BRAKE_POWER_PER_SECOND = 0.125;
	/** The maximum velocity of a shoot. */
	static final double MAX_SHOT_VELOCITY = 25.0;

	/** Static friction against the blades. */
	private static final double STATIC_FRICTION_SIDEWAYS = 4.0;
	/** Kinetic friction against the blades. */
	private static final double KINETIC_FRICTION_SIDEWAYS = 0.8;
	/** Static friction among the blades. */
	private static final double STATIC_FRICTION_STRAIGHT = 0.2;
	/** Kinetic friction among the blades. */
	private static final double KINETIC_FRICTION_STRAIGHT = 0.15;
	/** Compensation against friction for dashing in low update frequencies. */
	private static final double FRICTION_COMPENSATION = 9.8;

	/** Outline stroke style. */
  private static final Stroke STROKE = new BasicStroke(0.05f);
	/** Outline stroke style for position symbol. */
  private static final Stroke POSITION_STROKE = new BasicStroke(0.1f);
	/** The color of the helmet. */
	private Color helmetColor; 
	/** The color of the jersey. */
	private Color jerseyColor; 

	/** Player's name. */
	private String name;
	/** Player's number. */
	private int number;
	/** Does the player belong to the home team. */
	private boolean home;
	/** Player's position in the team. */
	private int position;

	/** The current power of the forward dash. */
	private double dashPower;
	/** The currect power of the backward dash. */
	private double brakePower;

  /** 
   *  Creates a player at location <code>p</code>. 
   */
  public Player(Point2 p)
    {
    super(MASS, RADIUS, new Vector2(0.0, 1.0), p);
    setTangentFriction(new Friction(STATIC_FRICTION_STRAIGHT,
    	KINETIC_FRICTION_STRAIGHT));
    setNormalFriction(new Friction(STATIC_FRICTION_SIDEWAYS,
    	KINETIC_FRICTION_SIDEWAYS));
		setName("");
		setNumber(0);
		setTeam(false);
		setPosition(0);
		setColors(Color.white, Color.blue);
    dashPower = 0.0;
    brakePower = 0.0;
    }


	/** Returns the player's name. */
	public String getName()
		{
		return name;
		}


	/** Sets the player's name to <code>s</code>. */
	public void setName(String s)
		{
		name = s;
		}


	/** Returns the player's number. */
	public int getNumber()
		{
		return number;
		}


	/** Sets the player's number to <code>n</code>. */
	public void setNumber(int n)
		{
		number = n;
		}


	/** Returns the player's position in the team. */
	public int getPlayPosition()
		{
		return position;
		}


	/** 
	 *	Set the player's position in the team to <code>p</code>.
	 *
	 *	@see Team#GOALKEEPER
	 *	@see Team#LEFT_DEFENSE
	 *	@see Team#RIGHT_DEFENSE
	 *	@see Team#LEFT_WING
	 *	@see Team#RIGHT_WING
	 *	@see Team#CENTER
	 */
	void setPosition(int p)
		{
		position = p;
		}


	/** Returns <code>true</code> if the player belongs to the home team. */
	public boolean isHome()
		{
		return home;
		}


	/** 
	 *	Sets the players team. If <code>h == true</code>, player 
	 *	is in the home team, otherwise in the visiting team.
	 */
	void setTeam(boolean h)
		{
		home = h;
		}

	
	/** Returns the player's helmet color. */
	public Color getHelmetColor()
		{
		return helmetColor;
		}


	/** Returns the player's jersey color. */
	public Color getJerseyColor()
		{
		return jerseyColor;
		}


	/** 
	 *	Sets the player's helmet color to <code>h</code> and
	 *	jersey color to <code>j</code>.
	 */
	public void setColors(Color h, Color j)
		{
	  helmetColor = h;
    jerseyColor = j;
		}


	/** 
	 *	Dashes the player forward with the proportional 
	 *	power of <code>p</code>. 
	 */
	public void dash(double p)
		{
		dashPower = MAX_DASH_POWER_PER_SECOND * p;
		brakePower = 0.0;
		}


	/** 
	 *	Brakes or dashes the player backward with the proportional 
	 *	power of <code>p</code>. 
	 */
	public void brake(double p)
		{
		brakePower = MAX_BRAKE_POWER_PER_SECOND * p;
		dashPower = 0.0;
		}


	/** 
	 *	Turns the player's orientation <code>a</code> 
	 *	radians clockwise.
	 */
	public void head(double a)
		{
//		Vector2 o = getOrientation();
		Vector2 o = new Vector2(0.0, 1.0);
		o.rotate(-a);
		setOrientation(o);
		}


	/**
	 *	Returns <code>true</code>, if player can reach the puck
	 *	<code>p</code>. 
	 */
	public boolean canReach(Puck p)
		{
    Vector2 d = new Vector2(getPosition(), p.getPosition());
    return (d.length() < (getRadius() + REACHABILITY + p.getRadius()));
    }


  /** 
   *  Advances the time <code>dt</code> seconds and updates the position, 
   *  velocity, and rotation of the player accordingly. 
   */
  public void advance(double dt)
    {
    super.advance(dt);		
    setAcceleration((Vector2)Vector2.ZERO.clone());
    if (dashPower > 0.0)
    	{
    	dashPower /= dt;
    	dashPower += (dt * FRICTION_COMPENSATION);
			Vector2 dashAcceleration = (Vector2)getOrientation().clone();
			dashAcceleration.multiply(dashPower);		
			setAcceleration(dashAcceleration);
			}
		if (brakePower > 0.0)
			{
    	brakePower /= dt;
    	brakePower += (dt * FRICTION_COMPENSATION);
			Vector2 brakeAcceleration = (Vector2)getOrientation().clone();
			brakeAcceleration.rotate180();
			brakeAcceleration.multiply(brakePower);		
			setAcceleration(brakeAcceleration);
			}
		dashPower = 0.0;
		brakePower = 0.0;
    }


  
  // ***************************************************
  // Methods defined in the Shapable interface  
  // ***************************************************

		public Shape shape()
			{
			double x = getPosition().getX();
			double y = getPosition().getY();    
			double r = getRadius();
			return new Ellipse2D.Double(x - r, y - r, r + r, r + r);
			}
		public boolean isDrawn() { return true; }
		public Stroke stroke() { return STROKE; }
		public Color color() { return helmetColor.darker(); }
		public boolean isFilled() { return true; }
		public Paint paint() { return helmetColor; }



	public ShapeIterator iterator()
		{
		Shapable[] s = new Shapable[4];


		s[0] = new ShapeAdapter() {
		  public Shape shape()
		    {
		    double x = getPosition().getX();
		    double y = getPosition().getY();    
		    double r = getRadius();
		    
		    return new Ellipse2D.Double(x - r - REACHABILITY, y - r - REACHABILITY, 
						r + r + 2.0 * REACHABILITY, r + r + 2.0 * REACHABILITY);
		    }
		  public Stroke stroke() {return STROKE;}
			public boolean isDrawn() {return true;}
			public Color color() {return Color.gray.brighter();}
			};

		s[1] = new ShapeAdapter() {
			public Shape shape()
				{
				double x = getPosition().getX();
				double y = getPosition().getY();    
				double r = getRadius() + REACHABILITY;

				GeneralPath result = null;

				switch (position)
					{
					case Position.GOALKEEPER:
						return new Line2D.Double(x - r, y, x + r, y);
					case Position.LEFT_DEFENSE:
						r /= Math.sqrt(2.0);
						return new Line2D.Double(x - r, y + r, x + r, y - r);
					case Position.RIGHT_DEFENSE:
						r /= Math.sqrt(2.0);
						return new Line2D.Double(x - r, y - r, x + r, y + r);
					case Position.LEFT_WING:
						double e = Math.PI / 10.0;
						double ar = Math.cos(Math.PI / 4.0 - e) * r;
						double br = Math.sin(Math.PI / 4.0 - e) * r;
						result = new GeneralPath();
						result.moveTo((float)(x - ar), (float)(y + br));
						result.lineTo((float)(x + br), (float)(y - ar));
						result.moveTo((float)(x - br), (float)(y + ar));
						result.lineTo((float)(x + ar), (float)(y - br));
						return result;					
					case Position.RIGHT_WING:
						e = Math.PI / 10.0;
						ar = Math.cos(Math.PI / 4.0 - e) * r;
						br = Math.sin(Math.PI / 4.0 - e) * r;
						result = new GeneralPath();
						result.moveTo((float)(x + ar), (float)(y + br));
						result.lineTo((float)(x - br), (float)(y - ar));
						result.moveTo((float)(x + br), (float)(y + ar));
						result.lineTo((float)(x - ar), (float)(y - br));
						return result;					
					case Position.CENTER:			
						e = Math.PI / 10.0;
						ar = Math.cos(e) * r;
						br = Math.sin(e) * r;
						result = new GeneralPath();
						result.moveTo((float)(x + br), (float)(y - ar));
						result.lineTo((float)(x + br), (float)(y + ar));
						result.moveTo((float)(x - br), (float)(y - ar));
						result.lineTo((float)(x - br), (float)(y + ar));
						return result;					
					}
				return null;
				}
			public boolean isDrawn() { return true; }
			public Stroke stroke() { return POSITION_STROKE; }
			public Color color() {return Color.gray.brighter();}
			};

		s[2] = new ShapeAdapter() {
		  public Shape shape()
		    {
		    double x = getPosition().getX();
		    double y = getPosition().getY();    
		    double r = getRadius();
		    Vector2 o = (Vector2)getOrientation().clone();
		    Vector2 q = (Vector2)getOrientation().clone();
		    o.multiply(r);
		    q.multiply(r + 1.0);
		    
		    float x1 = (float)(o.getDeltaX() + x);
		    float y1 = (float)(o.getDeltaY() + y);
				o.rotate90();
				q.rotate90();
		    float x2 = (float)(q.getDeltaX() + x);
		    float y2 = (float)(q.getDeltaY() + y);
				o.rotate90();
				q.rotate90();
		    float x3 = (float)(o.getDeltaX() + x);
		    float y3 = (float)(o.getDeltaY() + y);
				o.rotate90();
				q.rotate90();
		    float x4 = (float)(q.getDeltaX() + x);
		    float y4 = (float)(q.getDeltaY() + y);
				GeneralPath result = new GeneralPath();
				result.moveTo(x1, y1);
				result.quadTo(x2, y2, x3, y3);
				result.quadTo(x4, y4, x1, y1);
							
		    return result;		    
		    }
		  public Stroke stroke() {return STROKE;}
			public boolean isDrawn() {return true;}
			public boolean isFilled() {return true;}
			public Color color() {return jerseyColor.darker();}
			public Paint paint() {return jerseyColor;}
			};

		s[3] = new ShapeAdapter() {
		  public Shape shape()
		    {
		    double x = getPosition().getX();
		    double y = getPosition().getY();    
		    double r = getRadius();
		    Vector2 o = (Vector2)getOrientation().clone();
		    o.multiply(r + 0.5);
		    return new Line2D.Double(x, y, x + o.getDeltaX(), y + o.getDeltaY());
		    }
		  public Stroke stroke() {return POSITION_STROKE;}
			public boolean isDrawn() {return true;}
			public Color color() {return Color.black;}
			};

			
		
		return new ShapeIterator(s);
		}
  
  // ***************************************************




  public String toString()
    {
    return "Player: number = " + number + " name = " + name +
    	" home = " + home + " position = " + position + " " + super.toString();
    }  
  }
