/* 
 * Puck.java 1.0 2002-08-23
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
package fi.utu.cs.hockey;

import fi.utu.cs.physics.*;
import fi.utu.cs.appearance.*;
import fi.utu.cs.hockey.rink.Rink;
import fi.utu.cs.hockey.rink.FaceOff;

import java.awt.Shape;
import java.awt.Paint;
import java.awt.Color;
import java.awt.Stroke;
import java.awt.GradientPaint;
import java.awt.BasicStroke;
import java.awt.geom.Ellipse2D;
import java.util.Random;

/**
 *  A puck. 
 *
 *  @version 1.0, 020823
 *  @author  Jouni Smed
 */
public class Puck extends FrictionalBody implements MultiShapable, FaceOff
  {
	/** Diameter of the puck. */
	public static final double DIAMETER = 0.0762;
	/** Radius of the puck. */
  private static final double RADIUS = DIAMETER / 2.0;
  /** Mass of the puck in kilograms. */
  private static final double MASS = 0.170;
  /** The static and kinetic friction of the puck. */
  private static final Friction friction = new Friction(0.4, 0.2);
  
	/** The color of the puck. */	
  private static final Color DEFAULT_COLOR = Color.black;
  /** The stroke style for the outline. */
  private static final Stroke STROKE = new BasicStroke(0.01f);
  
  /** Range of velocity for the puck in a faceoff. */
  private static final double FACEOFF_RANGE = 1.5;
  /** Minimum velocity for the puck in a faceoff. */
  private static final double FACEOFF_MIN = 2.5;

	/** Faceoff positions. */
	private double[][] FACEOFF = {
		{ 0.0, 0.0 },		
		{ -Rink.FACE_OFF_OFFSET, -Rink.ENDZONE_FACE_OFF_OFFSET },
		{  Rink.FACE_OFF_OFFSET, -Rink.ENDZONE_FACE_OFF_OFFSET },
		{ -Rink.FACE_OFF_OFFSET, -Rink.NEUTRAL_ZONE_FACE_OFF_OFFSET },
		{  Rink.FACE_OFF_OFFSET, -Rink.NEUTRAL_ZONE_FACE_OFF_OFFSET },
		{ -Rink.FACE_OFF_OFFSET,  Rink.NEUTRAL_ZONE_FACE_OFF_OFFSET },
		{  Rink.FACE_OFF_OFFSET,  Rink.NEUTRAL_ZONE_FACE_OFF_OFFSET },
		{ -Rink.FACE_OFF_OFFSET,  Rink.ENDZONE_FACE_OFF_OFFSET },
		{  Rink.FACE_OFF_OFFSET,  Rink.ENDZONE_FACE_OFF_OFFSET } };

	/** Random number generator. */
	private Random random = new Random();

  /** 
   *  Creates a puck at position <code>p</code>. 
   */
  public Puck(Point2 p)
    {
    super(MASS, RADIUS, p, friction);
    }

	/** 
	 *	Drops the puck at faceoff spot <code>s</code>.
	 *
	 *	@see FaceOff
	 */
	public void faceoff(int s)
		{
		setPosition(FACEOFF[s][0], FACEOFF[s][1]);

		Vector2 v = new Vector2(1.0, 0.0);      		
		v.rotate(random.nextDouble() * 2 * Math.PI);
		v.multiply((random.nextDouble()) * FACEOFF_RANGE + FACEOFF_MIN);
		setVelocity(v);
		}
  
  // ***************************************************
  // Methods defined in the Shapable interface  
  // ***************************************************

  public Shape shape()
    {
    double x = getPosition().getX();
    double y = getPosition().getY();
    double r = getRadius();
    return new Ellipse2D.Double(x - r, y - r, 2 * r, 2 * r);
    }
    
  public boolean isDrawn()
    {
    return true;
    }
      
  public Stroke stroke()
    {
    return STROKE;
    }

  public Color color()
    {
    return DEFAULT_COLOR;
    }

  public boolean isFilled()
    {
    return true;
    }

  public Paint paint()
    {
    return DEFAULT_COLOR;
    }

	public ShapeIterator iterator()
		{
		Shapable[] s = new Shapable[1];

		s[0] = new ShapeAdapter() {
		  public Shape shape()
		    {
		    double x = getPosition().getX();
		    double y = getPosition().getY();    
		    double r = getRadius();
		    
		    return new Ellipse2D.Double(x - r - 0.35, y - r - 0.35, r + r + 0.7, r + r + 0.7);
		    }
			public boolean isFilled() {return true;}
			public Paint paint() {return Color.gray.brighter();}
			};

		return new ShapeIterator(s);
		}
  
  // ***************************************************

  public String toString()
    {
    return "Puck: " + super.toString();
    }  
  }
