/* 
 * Team.java 1.0 2002-09-26
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.hockey;

import fi.utu.cs.physics.Point2;
import fi.utu.cs.physics.Vector2;
import fi.utu.cs.hockey.rink.Rink;
import fi.utu.cs.hockey.rink.FaceOff;

import java.awt.Color;


/**
 *  A team. 
 *
 *  @version 1.0, 020926
 *  @author  Jouni Smed
 */

public class Team implements Position, FaceOff
  {
  /** Total number of players in a team. */
  public static final int PLAYERS = 6;

  private static final Vector2 HOME_DIRECTION = new Vector2(0.0, 1.0);

  private static final Vector2 VISITOR_DIRECTION = new Vector2(0.0, -1.0);

  /** Coordinates for a center spot faceoff. */
  private double CENTER_FACEOFF[][] = { 
    {  0.0,                                   // goalie
      -Rink.GOAL_LINE_OFFSET + Rink.GOAL_CREASE_RADIUS - 2.0 * Player.RADIUS},   
    {-Rink.FACE_OFF_OFFSET, 
     -Rink.BLUE_LINE_OFFSET - 2.0},    				// left defense
    { Rink.FACE_OFF_OFFSET, 
     -Rink.BLUE_LINE_OFFSET - 2.0},    				// right defense
    {-12.0,  -1.0},                           // left wing
    { 12.0,  -1.0},                           // right wing
    {  0.0,  -4.0 * Player.RADIUS} };         // center

  /** Coordinates for a defensive endzone spot faceoff. */
  private double DEFENSIVE_ENDZONE_FACEOFF[][] =  { 
    {-Rink.GOAL_WIDTH / 3.0,                        // goalie
     -Rink.GOAL_LINE_OFFSET + 2.0 * Player.RADIUS},          
    {-Rink.FACE_OFF_OFFSET - Rink.FACE_OFF_CIRCLE_RADIUS + 1.0, 
     -Rink.GOAL_LINE_OFFSET + 2.5},                 // left defense
    {-Rink.FACE_OFF_OFFSET + Rink.FACE_OFF_CIRCLE_RADIUS - 1.0, 
     -Rink.GOAL_LINE_OFFSET + 2.5},                 // right defense
    {-Rink.FACE_OFF_OFFSET - Rink.FACE_OFF_CIRCLE_RADIUS - 1.0, 
     -Rink.ENDZONE_FACE_OFF_OFFSET - 1.0},          // left wing
    { 0.0, -Rink.ENDZONE_FACE_OFF_OFFSET - 1.0},    // right wing
    {-Rink.FACE_OFF_OFFSET,                         // center
     -Rink.ENDZONE_FACE_OFF_OFFSET - (4.0 * Player.RADIUS)} }; 

  /** Coordinates for an offensive endzone spot faceoff. */
  private double OFFENSIVE_ENDZONE_FACEOFF[][] =  { 
    { Rink.GOAL_WIDTH / 4.0,                        // goalie
      -Rink.GOAL_LINE_OFFSET + Rink.GOAL_CREASE_RADIUS - 3.0 * Player.RADIUS},
    { Rink.FACE_OFF_OFFSET - Rink.FACE_OFF_CIRCLE_RADIUS + 1.0, 
      Rink.BLUE_LINE_OFFSET + 2.5},                 // left defense
    { Rink.FACE_OFF_OFFSET + Rink.FACE_OFF_CIRCLE_RADIUS - 1.0, 
      Rink.BLUE_LINE_OFFSET + 2.5},                 // right defense
    { 0.0, Rink.ENDZONE_FACE_OFF_OFFSET - 1.0},     // left wing
    { Rink.FACE_OFF_OFFSET + Rink.FACE_OFF_CIRCLE_RADIUS + 1.0, 
      Rink.ENDZONE_FACE_OFF_OFFSET - 1.0},          // right wing
    { Rink.FACE_OFF_OFFSET,                         // center
      Rink.ENDZONE_FACE_OFF_OFFSET - (4.0 * Player.RADIUS)} }; 

  /** Coordinates for a defensive neutral zone spot faceoff. */
  private double DEFENSIVE_NEUTRAL_ZONE_FACEOFF[][] =  { 
    {-Rink.GOAL_WIDTH / 4.0,                        // goalie
     -Rink.GOAL_LINE_OFFSET + 2.5 * Player.RADIUS},          
    {-Rink.FACE_OFF_OFFSET - Rink.FACE_OFF_CIRCLE_RADIUS + 1.0, 
     -Rink.BLUE_LINE_OFFSET - 5.0},                 // left defense
    {-Rink.FACE_OFF_OFFSET + Rink.FACE_OFF_CIRCLE_RADIUS - 1.0, 
     -Rink.BLUE_LINE_OFFSET - 5.0},                 // right defense
    {-Rink.FACE_OFF_OFFSET - Rink.FACE_OFF_CIRCLE_RADIUS - 1.0, 
     -Rink.NEUTRAL_ZONE_FACE_OFF_OFFSET - 1.0},          // left wing
    { 0.0, -Rink.NEUTRAL_ZONE_FACE_OFF_OFFSET - 1.0},    // right wing
    {-Rink.FACE_OFF_OFFSET,                         // center
     -Rink.NEUTRAL_ZONE_FACE_OFF_OFFSET - (4.0 * Player.RADIUS)} }; 

  /** Coordinates for an offensive neutral zone spot faceoff. */
  private double OFFENSIVE_NEUTRAL_ZONE_FACEOFF[][] =  { 
    { Rink.GOAL_WIDTH / 5.0,                        // goalie
      -Rink.GOAL_LINE_OFFSET + Rink.GOAL_CREASE_RADIUS - 2.5 * Player.RADIUS},
    { Rink.FACE_OFF_OFFSET - Rink.FACE_OFF_CIRCLE_RADIUS + 1.0, 
      -1.0},                                       // left defense
    { Rink.FACE_OFF_OFFSET + Rink.FACE_OFF_CIRCLE_RADIUS - 1.0, 
      -1.0},                                        // right defense
    { 0.0, Rink.NEUTRAL_ZONE_FACE_OFF_OFFSET - 1.0},     // left wing
    { Rink.FACE_OFF_OFFSET + Rink.FACE_OFF_CIRCLE_RADIUS + 1.0, 
      Rink.NEUTRAL_ZONE_FACE_OFF_OFFSET - 1.0},          // right wing
    { Rink.FACE_OFF_OFFSET,                         // center
      Rink.NEUTRAL_ZONE_FACE_OFF_OFFSET - (4.0 * Player.RADIUS)} }; 


  /** An array of the players. */ 
  private Player[] players;
  
  /** Team's name. */
  private String name;
  /** Is this the home team. */
  private boolean home;

  
  /** Constructs an empty, nameless team. */
  public Team()
    {
    players = new Player[PLAYERS];
    for (int i = 0; i < PLAYERS; i++)
      players[i] = null;
    setName(null);
    setTeam(false);
    } 


  /** Returns the team's name. */ 
  public String getName()
    {
    return name;
    }


  /** Sets the team's name to <code>n</code>. */  
  public void setName(String n)
    {
    name = n;
    }

  /** Returns <code>true</code>, if this is the home team. */ 
  public boolean isHome()
    {
    return home;
    }


  /** 
   *  Sets the team. If <code>h == true</code>, team is
   *  the home team, otherwise the visiting team.
   */
  void setTeam(boolean h)
    {
    home = h;
    }


  /**
   *  Creates, returns and adds a player to the position <code>p</code>.
   *  
   *  @see Position
   */
  public Player addPlayer(int p) 
    {
    Player player = new Player(faceoffLocation(p, CENTER_SPOT));
    player.setPosition(p);    
    player.setTeam(home);
    players[p] = player;
    return player;
    }

  
  /** Returns the player of the position <code>p</code>. */
  public Player getPlayer(int p)  
    {
    return players[p];
    }
  

  /**
   *  Returns the whether the position <code>p</code>
   *  is not taken already.
   */
  public boolean isFree(int p)
    {
    return (players[p] != null);
    }


  /**
   *  Moves the players of the team to faceoff locations
   *  when the faceoff is in the spot <code>s</code>.
   *
   *  @see FaceOff
   */
  public void faceoff(int s)
    {
    for (int i = 0; i < PLAYERS; i++)
      {
      if (players[i] != null) 
        {
        players[i].setPosition(faceoffLocation(i, s));
        players[i].setVelocity(0.0, 0.0);
        players[i].setAcceleration(0.0, 0.0);
        players[i].setOrientation((home ?
                      HOME_DIRECTION : VISITOR_DIRECTION));
        }
      }
    }


  /** 
   *  Returns faceoff position for player <code>p</code>
   *  in faceoff <code>s</code>.
   */
  private Point2 faceoffLocation(int p, int s)
    {
    if (s == CENTER_SPOT)
      return (home ? 
            new Point2(CENTER_FACEOFF[p][0], CENTER_FACEOFF[p][1]) :
            new Point2(-CENTER_FACEOFF[p][0], -CENTER_FACEOFF[p][1]));
    if (s == HOME_ENDZONE_LEFT)
      return (home ? 
            new Point2(DEFENSIVE_ENDZONE_FACEOFF[p][0],
                       DEFENSIVE_ENDZONE_FACEOFF[p][1]) :
            new Point2(-OFFENSIVE_ENDZONE_FACEOFF[p][0],
                       -OFFENSIVE_ENDZONE_FACEOFF[p][1]));
    if (s == HOME_ENDZONE_RIGHT)
      {
      int p2 = mirror(p);
      return (home ? 
            new Point2(-DEFENSIVE_ENDZONE_FACEOFF[p2][0], 
                        DEFENSIVE_ENDZONE_FACEOFF[p2][1]) :
            new Point2( OFFENSIVE_ENDZONE_FACEOFF[p2][0],
                       -OFFENSIVE_ENDZONE_FACEOFF[p2][1]));
      }
    if (s == VISITOR_ENDZONE_LEFT)
      {
      int p2 = mirror(p);
      return (home ? 
            new Point2(-OFFENSIVE_ENDZONE_FACEOFF[p2][0], 
                       OFFENSIVE_ENDZONE_FACEOFF[p2][1]) :
            new Point2( DEFENSIVE_ENDZONE_FACEOFF[p2][0],
                       -DEFENSIVE_ENDZONE_FACEOFF[p2][1]));
      }
    if (s == VISITOR_ENDZONE_RIGHT)
      return (home ? 
            new Point2(OFFENSIVE_ENDZONE_FACEOFF[p][0], 
                       OFFENSIVE_ENDZONE_FACEOFF[p][1]) :
            new Point2(-DEFENSIVE_ENDZONE_FACEOFF[p][0],
                       -DEFENSIVE_ENDZONE_FACEOFF[p][1]));
    if (s == HOME_NEUTRAL_ZONE_LEFT)
      return (home ? 
            new Point2(DEFENSIVE_NEUTRAL_ZONE_FACEOFF[p][0],
                       DEFENSIVE_NEUTRAL_ZONE_FACEOFF[p][1]) :
            new Point2(-OFFENSIVE_NEUTRAL_ZONE_FACEOFF[p][0],
                       -OFFENSIVE_NEUTRAL_ZONE_FACEOFF[p][1]));
    if (s == HOME_NEUTRAL_ZONE_RIGHT)
      {
      int p2 = mirror(p);
      return (home ? 
            new Point2(-DEFENSIVE_NEUTRAL_ZONE_FACEOFF[p2][0], 
                        DEFENSIVE_NEUTRAL_ZONE_FACEOFF[p2][1]) :
            new Point2( OFFENSIVE_NEUTRAL_ZONE_FACEOFF[p2][0],
                       -OFFENSIVE_NEUTRAL_ZONE_FACEOFF[p2][1]));
      }
    if (s == VISITOR_NEUTRAL_ZONE_LEFT)
      {
      int p2 = mirror(p);
      return (home ? 
            new Point2(-OFFENSIVE_NEUTRAL_ZONE_FACEOFF[p2][0], 
                       OFFENSIVE_NEUTRAL_ZONE_FACEOFF[p2][1]) :
            new Point2( DEFENSIVE_NEUTRAL_ZONE_FACEOFF[p2][0],
                       -DEFENSIVE_NEUTRAL_ZONE_FACEOFF[p2][1]));
      }
    if (s == VISITOR_NEUTRAL_ZONE_RIGHT)
      return (home ? 
            new Point2(OFFENSIVE_NEUTRAL_ZONE_FACEOFF[p][0], 
                       OFFENSIVE_NEUTRAL_ZONE_FACEOFF[p][1]) :
            new Point2(-DEFENSIVE_NEUTRAL_ZONE_FACEOFF[p][0],
                       -DEFENSIVE_NEUTRAL_ZONE_FACEOFF[p][1]));
    return null;
    }


	/** Returns the mirror position of position <code>p</code>. */
  private int mirror(int p)
    {
    switch (p)
      {
      case LEFT_DEFENSE: return RIGHT_DEFENSE;
      case RIGHT_DEFENSE: return LEFT_DEFENSE;
      case LEFT_WING: return RIGHT_WING;
      case RIGHT_WING: return LEFT_WING;
      }
    return p;
    }



	/** 
	 *	Returns <code>true</code>, if there exists a player
	 *	in the team that has a y-coordinate less
	 *	than <code>y</code>.
	 */
	boolean existsPlayerLessThan(double y)
		{
		for (int i = 0; i < PLAYERS; i++)
			{
			Player p = players[i];
			if (p.getPosition().getY() < y) return true;
			}
		return false;
		}

	/** 
	 *	Returns <code>true</code>, if there exists a player
	 *	in the team that has a y-coordinate greater
	 *	than <code>y</code>.
	 */
	boolean existsPlayerGreaterThan(double y)
		{
		for (int i = 0; i < PLAYERS; i++)
			{
			Player p = players[i];
			if (p.getPosition().getY() > y) return true;
			}
		return false;
		}

	/** 
	 *	Returns <code>true</code>, if all players in 
	 *	in the team have y-coordinate less
	 *	than <code>y</code>.
	 */
	boolean forAllPlayersLessThan(double y)
		{
		for (int i = 0; i < PLAYERS; i++)
			{
			Player p = players[i];
			if (p.getPosition().getY() > y) return false;
			}
		return true;
		}

	/** 
	 *	Returns <code>true</code>, if all players in 
	 *	in the team have y-coordinate greater
	 *	than <code>y</code>.
	 */
	boolean forAllPlayersGreaterThan(double y)
		{
		for (int i = 0; i < PLAYERS; i++)
			{
			Player p = players[i];
			if (p.getPosition().getY() < y) return false;
			}
		return true;
		}


  
  public String toString()
    {
    String result = "Team: " + name + " (" 
                    + (home ? "home" : "visitor") + ")\n";
    result = result + "Goalkeeper: " 
                    + players[GOALKEEPER] + "\n";
    result = result + "Left defense: " 
                    + players[LEFT_DEFENSE] + "\n";
    result = result + "Right defense: " 
                    + players[RIGHT_DEFENSE] + "\n";
    result = result + "Left wing: " 
                    + players[LEFT_WING] + "\n";
    result = result + "Right wing: " 
                    + players[RIGHT_WING] + "\n";
    result = result + "Center: " 
                    + players[CENTER] + "\n";
    return result;
    }
  }