/* 
 * AI.java 1.0 2002-11-21
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.hockey.ai;

import fi.utu.cs.hockey.net.Communication;
import fi.utu.cs.hockey.net.Client;

import java.awt.Color;
import java.util.NoSuchElementException;

/**
 *	AI for a hockey player.
 *
 *  @version 1.0, 021121
 *  @author  Jouni Smed
 */

public abstract class AI extends Client 
	{	
	/** Whether the player quits, if it receives a game over message. */
	private boolean neverQuit = false;
	
	/** 
	 *  The amount of successive game over signals that must be received
	 *	before the player AI quits.
	 */
	private static final int maxGameOverCount = 5;
	
	/* The current successive game over signals count. */
	private int gameOverCount = 0;
	
	/**
	 *	A class for iterating player positions.
	 */
	public static class PlayerIterator 
		{
		private int finger;
		private int last;
		
		/**
		 *	Constructs an iterator from position <code>s</code>
		 *	to position <code>e</code>.
		 */
		PlayerIterator(int s, int e)
			{
			finger = s;
			last = e;
			}
		
		/**
		 *	Returns <code>true</code>, if the positions are not exhausted.
		 */
		public boolean hasNext()
			{
			return (finger <= last);
			}

		/**
		 *	Returns the next position.
		 *
		 *	@throws NoSuchElementException positions are exhausted.
		 */
		public int next() 
			{
			if (!hasNext()) 
				throw new NoSuchElementException("Player positions exhausted");
			int result = finger;
			finger++;
			return result;
			}
		}


	/** Creates a player AI */
	protected AI()
		{
		super();
		}

	/**
	 *	Creates a player AI with name <code>s</code>,
	 *	team name <code>t</code>, number <code>n</code>,
	 *	position <code>p</code>, helmet color <code>h</code>,
	 *	and jersey color <code>j</code>.
	 */
	protected AI(String s, String t, int n, int p, Color h, Color j)
		{
		super(s, t, n, p, h, j);
		}

	/**
	 *	Sets that the AI does not quit even if it receives 
	 *	the game over signal.
	 */
	public final void setNeverQuit()
		{
		neverQuit = true;
		}

	/**
	 *	Calls repeatedly the <code>react</code> method.
	 */
	public final void run()
		{
		while (neverQuit || (gameOverCount < maxGameOverCount))
			{
			try 
				{
				react();
				}
			catch (Exception e)
				{
				System.err.println(e);
				System.exit(1);
				}
			yield();
			if (isGameOver()) gameOverCount++;
			else gameOverCount = 0;
			}
		}
		
	/**
	 *	Reacts to the game events.
	 *	This method connects the player AI to the game engine.
	 */
	protected abstract void react();


	/**
	 *	Returns <code>true</code> if the puck is within the
	 *	reach of the player.
	 */
	public boolean isPuckWithinReach()
		{
		double[] me = player(true, getPosition());
		double[] puck = puck();
		double dx = puck[X_COORDINATE] - me[X_COORDINATE];
		double dy = puck[Y_COORDINATE] - me[Y_COORDINATE];		
		double d = Math.sqrt(dx * dx + dy * dy);		
    return (d < (Constants.PLAYER_RADIUS + Constants.PLAYER_REACHABILITY 
    							+ (Constants.PUCK_DIAMETER / 2.0)));		
		}


	/**
	 *	Returns the angle of heading to the point 
	 *	(<code>x</code>, <code>y</code>) from the player's current
	 *	location.
	 *
	 *	@see Communication#head(double)
	 */
	public double headingTo(double x, double y)
		{
		double[] me = player(true, getPosition());
		
		double dx = x - me[X_COORDINATE];
		double dy = y - me[Y_COORDINATE];		
		double h = Math.sqrt(dx * dx + dy * dy);

		if (h == 0.0) return 0.0;
		double result = Math.acos(dy / h);
		
		if (dx >= 0.0) return result;
		else return -result;		
		}

	/**	
	 *	Returns the angle of heading to the point 
	 *	(<code>p[X_COORDINATE]</code>, <code>p[Y_COORDINATE]</code>) 
	 *	from the player's current location.
	 *
	 *	@see Communication#head(double)
	 *	@see Communication#X_COORDINATE
	 *	@see Communication#Y_COORDINATE
	 */
	public double headingTo(double[] p)
		{
		return headingTo(p[X_COORDINATE], p[Y_COORDINATE]);
		}

	/** Returns an iterator for all player positions. */
	public PlayerIterator iteratorAll()
		{
		return new PlayerIterator(GOALKEEPER, CENTER);
		}
	
	/** Returns an iterator for defense positions. */
	public PlayerIterator iteratorDefense()
		{
		return new PlayerIterator(LEFT_DEFENSE, RIGHT_DEFENSE);
		}
		
	/** Returns an iterator for forward positions. */
	public PlayerIterator iteratorForward()
		{
		return new PlayerIterator(LEFT_WING, CENTER);
		}	
	}

