/* 
 * Client.java 1.0 2002-08-21
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.hockey.net;

import fi.utu.cs.hockey.Team;

import java.net.InetAddress;
import java.net.Socket;
import java.net.MulticastSocket;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.UnknownHostException;
import java.net.SocketException;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.BufferedWriter;
import java.io.OutputStreamWriter;
import java.util.Random;
import java.awt.Color;


/**
 *	A hockey player client.
 *
 *  @version 1.0, 020821
 *  @author  Jouni Smed
 */

public abstract class Client extends Thread implements Communication
	{
	/** The range in milliseconds for delaying outgoing messages. */
	private final int UPDATE_DELAY_RANGE = 6;

	/** Command offset for player identification. */
	static final int ID				= 0;
	/** Command offset for player's shoot power. */
	static final int SHOOT 		= 1;
	/** Command offset for player's dash power. */
	static final int DASH			= 2;
	/** Command offset for player's heading. */
	static final int HEADING	= 3;
	/** Command offset for player's message. */
	static final int SAY			= 4;
	/** Total number of commands. */
	static final int COMMANDS	= 5;
	
	/** Output buffer size. */
	private final int OUTPUT_BUFFER_SIZE = COMMANDS * 8;
	/** Input buffer size. */
	private final int INPUT_BUFFER_SIZE = Server.MESSAGES * 8;

	/** Keep puck command (i.e., keeping is negative shooting). */
	private static final double KEEP_PUCK = -1.0;

	/** Server address. */
	private InetAddress server;
	/** Multicast group address. */
	private InetAddress group;
	/** Server port for commands. */
	private int serverPort;
	/** Server port for initialization. */
	private int initializationPort;
	/** Multicast port. */
	private int groupPort;
	/** Socket for commands. */
	private DatagramSocket outputSocket;
	/** Socket for multicasted messages. */
	private MulticastSocket inputSocket;
	/** Packet for command output. */
	private DatagramPacket outputPacket;
	/** Packet for message input. */
	private DatagramPacket inputPacket;
	/** Online output for commands. */
	private long[] output = new long[COMMANDS];
	/** Offline output for preparing the packet. */
	private long[] output2 = new long[COMMANDS];
	/** Online input for messages. */
	private long[] input = new long[Server.MESSAGES];
	/** Offline input for extracting from the packet. */
	private long[] input2 = new long[Server.MESSAGES];
	/** Buffer for output packets. */
	private byte[] outputBuffer = new byte[OUTPUT_BUFFER_SIZE];
	/** Buffer for input packets. */
	private byte[] inputBuffer = new byte[INPUT_BUFFER_SIZE];

	/** The name of the player. */
	private String name;
	/** The name of the player's team. */
	private String team;
	/** The player's number. */
	private int number;
	/** Whether the player is in the home team. */
	private boolean home;
	/** Position of the player in the team. */
  private int position;
  /** Color of the player's helmet. */
  private Color helmetColor;
  /** Color of the player's jersey. */
  private Color jerseyColor;
  /** Is player joined in a game. */
  private boolean joined;


	/**
	 *	Creates a player client. 
	 */
	protected Client()
		{
		this("", "", 1, 0, Color.blue, Color.white);
		}

  
	/**
	 *	Creates a player client with name <code>s</code>,
	 *	team name <code>t</code>, number <code>n</code>,
	 *	position <code>p</code>, helmet color <code>h</code>,
	 *	and jersey color <code>j</code>.
	 */
	protected Client(String s, String t, int n, int p, Color h, Color j)
		{
		setPlayerName(s);
		setTeamName(t);
		setNumber(n);
		setPosition(p);
		setHelmetColor(h);
		setJerseyColor(j);
		joined = false;
		setTeam(false);
		}

	
	/**
	 *	Cleanup for the garbage collector.
	 */
	public void finalize() throws Throwable
		{
		try
			{
			if (outputSocket != null) outputSocket.close();
			if (inputSocket != null) 
				{
				inputSocket.leaveGroup(group);
				inputSocket.close();
				}
			}
		finally
			{
			super.finalize();
			}
		}


	/** Sets whether the player is in the home team. */
	final void setTeam(boolean h)
		{
		home = h;
		// Home has positive IDs, visitor negative.
		output[ID] = (home ? 1L : -1L) * (long)(position + 1);
		output2[ID] = (home ? 1L : -1L) * (long)(position + 1);
		}


	/**
	 *	Connects to a game server.
	 *	The server has two ports, one for initialization upon joining and
	 *	one for client input throughout the game. The server sends
	 *	output to a multicast group. Hence, parameter <code>gAddr</code>
	 *	must be a D-class address (i.e., within the range
	 *  224.0.0.0-239.255.255.255). Preferably, it should be
	 *  from the range 239.0.0.0-239.255.255.255 which is intended
	 *  for local area network (LAN) use only.
	 *	
	 *	@param sAddr Server address
	 *	@param sPort Server port for client input
	 *	@param iPort Server port for client initialization
	 *	@param gAddr Multicast group for server output
	 *	@param gPort Multicast port
	 *
	 */	
	public void connect(String sAddr, int sPort, int iPort,
							 				String gAddr, int gPort) 
		{
		try
			{
			server = InetAddress.getByName(sAddr);		
			}
		catch (UnknownHostException e)
			{
			throw new RuntimeException("Unknown server address " + sAddr +  ": " + e);
			}
		serverPort = sPort;
		initializationPort = iPort;

		try
			{
			group = InetAddress.getByName(gAddr); 
			}
		catch (UnknownHostException e)
			{
			throw new RuntimeException("Unknown address " + gAddr +  ": " + e);
			}

		if (!group.isMulticastAddress())
			throw new IllegalArgumentException(gAddr + 
				" is not a multicast address.");
		groupPort = gPort;

		try
			{
			outputSocket = new DatagramSocket();
			}
		catch (SocketException e)
			{
			throw new RuntimeException("Socket problem: " + e);
			}

		try
			{
			inputSocket = new MulticastSocket(groupPort);
			inputSocket.joinGroup(group);
			}
		catch (IOException e)
			{
			throw new RuntimeException("Multicast socket problem: " + e);
			}

		outputPacket = new DatagramPacket(outputBuffer, outputBuffer.length, 
																			server, serverPort);
		inputPacket = new DatagramPacket(inputBuffer, inputBuffer.length);

		initializeConnection();				
		new ConnectionHandler().start();
		joined = true;
		}


	/** Connects the server for initialization. */
	private void initializeConnection() 
		{
		Socket socket = null;
		try
			{
			socket = new Socket(server, initializationPort);
			BufferedReader in = new BufferedReader(new 
													InputStreamReader(socket.getInputStream()));
			PrintWriter out = new PrintWriter(new BufferedWriter(new
												OutputStreamWriter(socket.getOutputStream())), true); 
			
			// Sends the server:
			// (1) player's name
			// (2) player's number
			// (3) player's position
			// (4) team's name
			// (5) team's helmet color
			// (6) team's jersey color
			out.println(name);
			out.println(number);
			out.println(position);
			out.println(team);
			out.println(helmetColor.getRGB());
			out.println(jerseyColor.getRGB());
			out.flush();

			String s = in.readLine();
			
			if (s.equals(Server.INIT_ACCEPTED_HOME))
				{
				setTeam(true);
				}
			else if (s.equals(Server.INIT_ACCEPTED_VISITOR))
				{
				setTeam(false);
				}
			else
				{
				throw new RuntimeException("Cannot join the game: " + s);
				}
			}
		catch (IOException e) 
			{
			throw new RuntimeException("Initialization: socket failed. " + e);
			}
		finally
			{
			try	
				{
				if (socket != null) socket.close();
				}
			catch (IOException e) { }
			}
		}


	/** Copies the outputs to the packet buffer. */
	private void outputToBuffer()
		{
		switchOutputs();
		for (int i = 0; i < output2.length; i++)
			{
			long value = output2[i];
			// Keep the ID and the HEADING but clear the rest.
			if (i != ID & i != HEADING) output2[i] = 0L;
			for (int j = ((i + 1) * 8) - 1; j >= (i * 8); j--)
				{
				outputBuffer[j] = (byte)value;
				value >>= 8;
				}
			}
		}


	/** 
	 *	Copies the packet buffer to the inputs or returns 
	 *	<code>false</code> if message is not proper. 
	 */
	private boolean bufferToInput()
		{
		for (int i = 0; i < input2.length; i++)
			{
			long value = (long)inputBuffer[i * 8];
			value &= 0xFF;
			for (int j = ((i * 8) + 1); j < ((i + 1) * 8); j++)
				{
				value <<= 8;
				long nextByte = (long)inputBuffer[j];
				nextByte &= 0xFF;
				value |= nextByte;

				}
			if ((i == 0) && ((value & Server.UNUSED_BITS) != 0L))
				return false;
			input2[i] = value;
			}
		switchInputs();
		return true;
		}


	/** Switches the online and offline outputs. */
	private void switchOutputs()
		{
		long[] temp = output;
		synchronized(this)
			{
			output = output2;
			}
		output2 = temp;
		}


	/** Switches the online and offline inputs. */
	private void switchInputs()
		{
		long[] temp = input;
		synchronized(this)
			{
			input = input2;
			}
		input2 = temp;
		}


	/**
	 *	A thread which handles the traffic between
	 *	the client and the server.
	 */
	private class ConnectionHandler extends Thread
		{
		/** A random number generator. */
		private Random random = new Random();	
		
		/**	It's a daemon, which dies when client dies. */
		public ConnectionHandler()
			{
			setDaemon(true);
			}

		/**
		 *	Handles the traffic. The loop:
		 *	(1) receive a packet from the server,
		 *	(2) copy the packet contents to the inputs,
		 *	(3) copy the outputs to a packet,
		 *	(4) sleep a random time, and
		 *	(5) send the packet to the server.
		 */
		public void run()
			{
			while (true)
				{
				try
					{
					inputSocket.receive(inputPacket);
					if (bufferToInput())
						{
						outputToBuffer();
						try
							{
							sleep(random.nextInt(UPDATE_DELAY_RANGE) + 1);
							}
						catch (InterruptedException e) { }
						outputSocket.send(outputPacket);
						}
					else
						{
						System.err.println("Received an improper packet originated from "
																+ inputPacket.getAddress());
						}
					}
				catch (IOException e)	
					{
					throw new RuntimeException("Connection failure: " + e);
					}
				}
			}
		}



/* ************************************************************************
 *
 * Implementations for the methods defined in the Communication interface.
 * 
 * ************************************************************************
 */

	public final String getPlayerName()
		{
		return name;
		}


	public final void setPlayerName(String s)
		{
		if (!joined) name = s;
		}


	public final String getTeamName()
		{
		return team;
		}


	public final void setTeamName(String t)
		{
		if (!joined) team = t;
		}

	public final int getNumber()
		{
		return number;
		}


	public final void setNumber(int n)
		{
		if (n <= 0 || n > 99) 
			throw new IllegalArgumentException("Player number " + n + 
				" is outside the range [1, 99].");
		if (!joined) number = n;
		}


	public final int getPosition()
		{
		return position;
		}


	public final void setPosition(int p)
		{
		if (p < GOALKEEPER || p > CENTER) 
			throw new IllegalArgumentException("Player position illegal.");
		if (!joined) position = p;
		}


	public final Color getHelmetColor()
		{
		return helmetColor;
		}


	public final void setHelmetColor(Color h)
		{
		if (!joined) helmetColor = h;
		}


	public final Color getJerseyColor()
		{
		return jerseyColor;
		}


	public final void setJerseyColor(Color j)
		{
		if (!joined) jerseyColor = j;
		}


	public final boolean isInHomeTeam()
		{
		return home;
		}


	public final boolean isJoined()
		{
		return joined;
		}


/* ******************************************************************** */

	public final void shoot(double p)
		{
		if (p < 0.0 || p > 1.0) 
			throw new IllegalArgumentException("Shoot: argument value " 
				+ p + " is outside the range [0.0, 1.0].");
		synchronized(this)
			{
			output[SHOOT] = Double.doubleToRawLongBits(p); 
			}
		}


	public final void dash(double p)
		{
		if (p < 0.0 || p > 1.0) 
			throw new IllegalArgumentException("Dash: argument value " 
				+ p + " is outside the range [0.0, 1.0].");
		synchronized(this)
			{
			output[DASH] = Double.doubleToRawLongBits(p);
			}
		}


	public final void brake(double p)
		{
		if (p < 0.0 || p > 1.0) 
			throw new IllegalArgumentException("Brake: argument value " 
				+ p + " is outside the range [0.0, 1.0].");
		synchronized(this)
			{
			output[DASH] = Double.doubleToRawLongBits(-p);
			}
		}


	public final void head(double a)
		{
		synchronized(this)
			{
			output[HEADING] = Double.doubleToRawLongBits(a);
			}
		}


	public final void say(long s)
		{
		synchronized(this)
			{
			output[SAY] = s;
			}	
		}


	public final void keepPuck()
		{
		if (position != GOALKEEPER) 
			throw new RuntimeException("Keeping the puck: not a goalie.");
		synchronized(this)
			{
			output[SHOOT] = Double.doubleToRawLongBits(KEEP_PUCK);
			}
		}


/* ******************************************************************** */


	public final long time()
		{
		return input[Server.TIME];
		}

		
	public final boolean isGameStopped()
		{
		return ((input[Server.REFEREE] & Server.GAME_STOPPED) != 0);
		}

	
	public final boolean isGameOver()
		{
		return (input[Server.REFEREE] == Server.GAME_OVER);
		}

		
	public final int score(boolean us)
		{
		return (!(home ^ us) ? 
						(int)(input[Server.SCORE] >>> 32) : 
						(int)input[Server.SCORE]);
		}


	public final double[] puck()
		{
		double[] result = new double[2];
		synchronized(this)
			{
			result[0] = (home ? 1.0 : -1.0) * 
									Double.longBitsToDouble(input[Server.PUCK_X]);
			result[1] = (home ? 1.0 : -1.0) * 
									Double.longBitsToDouble(input[Server.PUCK_Y]);
			}
		return result;
		}


	public final double[] player(boolean us, int p)
		{
		if (p < GOALKEEPER || p > CENTER) 
			throw new IllegalArgumentException("Player index " + p + 
				" out of bounds.");
		double[] result = new double[3];
		int base = (!(home ^ us) ? 
								Server.HOME : Server.VISITOR) + (p * Server.OFFSET_SIZE);
		synchronized(this)
			{
			result[0] = (home ? 1.0 : -1.0) * 
									Double.longBitsToDouble(input[base + Server.OFFSET_X]);
			result[1] = (home ? 1.0 : -1.0) * 
									Double.longBitsToDouble(input[base + Server.OFFSET_Y]);
			result[2] = (home ? 0.0 : Math.PI) + 
									Double.longBitsToDouble(input[base + Server.OFFSET_ANGLE]);
			}
		return result;
		}


	public final long[] messages(boolean us)
		{
		long[] result = new long[6];
		int base = (!(home ^ us) ? Server.HOME : Server.VISITOR) 
			+ Server.OFFSET_MSG;
		synchronized(this)
			{
			for (int i = 0; i < 6; i++)
				{
				result[i] = input[base];
				base += Server.OFFSET_SIZE;
				}
			}
		return result;
		}
	
	
	public final int faceoff()
		{
		long referee = input[Server.REFEREE];
		int result = NO_FACEOFF;

		if ((referee & Server.FACEOFF_CENTER) != 0) return CENTER_SPOT;
		
		if ((referee & Server.FACEOFF_HOME_ENDZONE_LEFT) != 0) 
			return (home ? OUR_ENDZONE_LEFT : THEIR_ENDZONE_RIGHT);
		if ((referee & Server.FACEOFF_HOME_ENDZONE_RIGHT) != 0) 
			return (home ? OUR_ENDZONE_RIGHT : THEIR_ENDZONE_LEFT);
			
		if ((referee & Server.FACEOFF_VISITOR_ENDZONE_LEFT) != 0) 
			return (!home ? OUR_ENDZONE_RIGHT : THEIR_ENDZONE_LEFT);
		if ((referee & Server.FACEOFF_VISITOR_ENDZONE_RIGHT) != 0) 
			return (!home ? OUR_ENDZONE_LEFT : THEIR_ENDZONE_RIGHT);
			
		if ((referee & Server.FACEOFF_HOME_NEUTRAL_ZONE_LEFT) != 0) 
			return (home ? OUR_NEUTRAL_ZONE_LEFT : THEIR_NEUTRAL_ZONE_RIGHT);
		if ((referee & Server.FACEOFF_HOME_NEUTRAL_ZONE_RIGHT) != 0) 
			return (home ? OUR_NEUTRAL_ZONE_RIGHT : THEIR_NEUTRAL_ZONE_LEFT);
			
		if ((referee & Server.FACEOFF_VISITOR_NEUTRAL_ZONE_LEFT) != 0) 
			return (!home ? OUR_NEUTRAL_ZONE_RIGHT : THEIR_NEUTRAL_ZONE_LEFT);
		if ((referee & Server.FACEOFF_VISITOR_NEUTRAL_ZONE_RIGHT) != 0) 
			return (!home ? OUR_NEUTRAL_ZONE_LEFT : THEIR_NEUTRAL_ZONE_RIGHT);
		return result;
		}
		
	
	public final boolean isGoal(boolean us)
		{
		long bit = (!(home ^ us) ? Server.HOME_GOAL : Server.VISITOR_GOAL);
		return ((input[Server.REFEREE] & bit) != 0);
		}
	

	public final boolean isIcing(boolean us)
		{
		long bit = (!(home ^ us) ? Server.HOME_ICING : Server.VISITOR_ICING);
		return ((input[Server.REFEREE] & bit) != 0);
		}
	

	public final boolean isOffside(boolean us)
		{
		long bit = (!(home ^ us) ? Server.HOME_OFFSIDE : Server.VISITOR_OFFSIDE);
		return ((input[Server.REFEREE] & bit) != 0);
		}


	public final boolean isInterferingGoalie(boolean us)
		{
		long bit = (!(home ^ us) ? Server.HOME_INTERFERING_GOALIE : 
											Server.VISITOR_INTERFERING_GOALIE);
		return ((input[Server.REFEREE] & bit) != 0);
		}


	public final boolean isGoalieKeepingPuck(boolean us)
		{
		long bit = (!(home ^ us) ? Server.HOME_GOALIE_HAS_PUCK : 
											Server.VISITOR_GOALIE_HAS_PUCK);
		return ((input[Server.REFEREE] & bit) != 0);
		}


	public final boolean isFoul(boolean us)
		{
		long bit = (!(home ^ us) ? Server.HOME_OTHER : Server.VISITOR_OTHER);
		return ((input[Server.REFEREE] & bit) != 0);
		}
	

	public final boolean isPuckUnplayable()
		{
		return ((input[Server.REFEREE] & Server.PUCK_UNPLAYABLE) != 0);
		}
		
	}