/* 
 * Communication.java 1.0 2002-08-21
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.hockey.net;

import fi.utu.cs.hockey.Team;
import fi.utu.cs.hockey.rink.FaceOff;
import fi.utu.cs.hockey.Position;

import java.awt.Color;

/**
 *	Communication between a hockey player client and server from the 
 *	client's perspective. 
 *	The client has six methods for affecting the server: 
 *	<code>shoot</code>, <code>dash</code>, <code>brake</code>,
 *	<code>head</code>, <code>say</code>, and <code>keepPuck</code>.
 *	On the other hand, the client receives from the server,
 *	in addition to the state of the game, the location
 *	of the puck, the location, orientation and messages from 
 *	all players.
 *
 *  @version 1.0, 020821
 *  @author  Jouni Smed
 */

public interface Communication extends Position
	{	
	/** No faceoff. */
	public int NO_FACEOFF								= 0;
	/** Faceoff at the center spot. */
	public int CENTER_SPOT							= 1;
	/** Faceoff at our endzone on the left side. */	
	public int OUR_ENDZONE_LEFT					= 2;
	/** Faceoff at our endzone on the right side. */	
	public int OUR_ENDZONE_RIGHT				= 3;
	/** Faceoff at our neutral zone on the left side. */	
	public int OUR_NEUTRAL_ZONE_LEFT		= 4;
	/** Faceoff at our neutral zone on the right side. */	
	public int OUR_NEUTRAL_ZONE_RIGHT		= 5;
	/** Faceoff at their neutral zone on the left side. */	
	public int THEIR_NEUTRAL_ZONE_LEFT	= 6;
	/** Faceoff at their neutral zone on the right side. */	
	public int THEIR_NEUTRAL_ZONE_RIGHT	= 7;
	/** Faceoff at their endzone on the left side. */	
	public int THEIR_ENDZONE_LEFT				= 8;
	/** Faceoff at their endzone on the right side. */	
	public int THEIR_ENDZONE_RIGHT			= 9;

	/** Array index for the x-coordinate. */
	public int X_COORDINATE 	= 0;
	/** Array index for the y-coordinate. */
	public int Y_COORDINATE 	= 1;
	/** Array index for the angle of heading. */
	public int HEADING_ANGLE	= 2;
	
	/** Returns the player's name. */
	public String getPlayerName();

	/** 
	 * 	Sets the player's name to <code>s</code>. 
	 * 	After the player has joined a game, 
	 * 	this method will have no effect.
	 *
	 *	@see #isJoined()
	 */
	public void setPlayerName(String s);

	/** Returns the name of the player's team. */
	public String getTeamName();

	/** 
	 * 	Sets the name of the player's team to <code>t</code>. 
	 * 	After the player has joined a game, 
	 * 	this method will have no effect.
	 *
	 *	@see #isJoined()
	 */
	public void setTeamName(String t);

	/** Returns the player's number. */
	public int getNumber();
	
	/** 
	 * 	Sets the player's number to <code>n</code>. 
	 * 	The number should be from the range 1-99.
	 * 	After the player has joined a game, 
	 * 	this method will have no effect.
	 *
	 *	@see #isJoined()
	 */
	public void setNumber(int n);
	
	/** 
	 *	Returns the player's position in the team.
	 *
	 *	@see #GOALKEEPER
	 *	@see #LEFT_DEFENSE
	 *	@see #RIGHT_DEFENSE
	 *	@see #LEFT_WING
	 *	@see #RIGHT_WING
	 * 	@see #CENTER
	 */
	public int getPosition();
	
	/** 
	 *	Sets the player's position in the team to <code>p</code>.
	 * 	After the player has joined a game, 
	 * 	this method will have no effect.
	 *
	 *	@see #isJoined()
	 *
	 *	@see #GOALKEEPER
	 *	@see #LEFT_DEFENSE
	 *	@see #RIGHT_DEFENSE
	 *	@see #LEFT_WING
	 *	@see #RIGHT_WING
	 * 	@see #CENTER
	 */
	public void setPosition(int p);

	/** Returns the color of the player's helmet. */
	public Color getHelmetColor();

	/** 
	 * 	Sets the name of the player's helmet to <code>h</code>. 
	 * 	After the player has joined a game, 
	 * 	this method will have no effect.
	 *
	 *	@see #isJoined()
	 */
	public void setHelmetColor(Color h);

	/** Returns the color of the player's jersey. */
	public Color getJerseyColor();

	/** 
	 * 	Sets the name of the player's jersey to <code>j</code>. 
	 * 	After the player has joined a game, 
	 * 	this method will have no effect.
	 *
	 *	@see #isJoined()
	 */
	public void setJerseyColor(Color j);


	/** 
	 *	Returns <code>true</code> if the player 
	 *	belongs to the home team; otherwise returns
	 *  <code>false</code>. 
	 *	Before the player has joined a game,
	 *	this method returns only <code>false</code>.
	 *
	 *	@see #isJoined()
	 */
	public boolean isInHomeTeam();

	/**
	 *	Returns <code>true</code>, if the player has
	 *	joined a game. After joining, changes in
	 *	name, number, position or colors have no effect.
	 *
	 *	@see #setPlayerName(String)
	 *	@see #setTeamName(String)
	 *	@see #setNumber(int)
	 *	@see #setPosition(int)
	 *	@see #setHelmetColor(Color)
	 *	@see #setJerseyColor(Color)
	 */
	public boolean isJoined();
	
	
	/**
	 *	Shoots the puck with the relative power <code>p</code>,
	 *	where <code>p</code> is from the range [0.0, 1.0].
	 */
	public void shoot(double p);
		
	/**
	 *	Dashes the player with the relative power <code>p</code>,
	 *	where <code>p</code> is from the range [0.0, 1.0].
	 *	Methods <code>dash</code> and <code>brake</code>
	 *	cancel out each other (i.e., only the last call
	 *	will remain active).
	 *	
	 *	@see #brake(double)
	 */
	public void dash(double p);

	/**
	 *	Brakes (or dashes backwards) the player with the 
	 *  relative power <code>p</code>, where <code>p</code> 
	 *	is from the range [0.0, 1.0].
	 *	Methods <code>dash</code> and <code>brake</code>
	 *	cancel out each other (i.e., only the last call
	 *	will remain active).
	 *	
	 *	@see #dash(double)
	 */
	public void brake(double p);

	/**
	 *	Sets the player's heading to the angle <code>a</code> in radians.
	 */
	public void head(double a);

	/**
	 *	Says the 64-bit long word <code>s</code>.
	 *	The message is received both by the own team
	 *	and the opponent.
	 */
	public void say(long s);

	/**
	 *	Keeps the puck, if the player is a goalkeeper.
	 *	Cancels out any calls to the method <code>shoot</code>.
	 *
	 *	@see #shoot(double)
	 */
	public void keepPuck();


	/** Returns the time left in the game in milliseconds. */
	public long time();

	/** Returns whether the game is stopped. */
	public boolean isGameStopped();
	
	/** Returns whether the game is over. */
	public boolean isGameOver();

	/** 
	 *	Returns the score of a team. 
	 *	If <code>us == true</code>, the team is own, 
	 *	otherwise the team is opponent.
	 */
	public int score(boolean us);

	/** 
	 *	Returns the location of the puck. 
	 *	
	 *	@see #X_COORDINATE
	 *	@see #Y_COORDINATE
	 */
	public double[] puck();

	/** 
	 *	Returns the location and the angle of the heading of a player.
	 *	If <code>us == true</code>, the player is in the own
	 *	team, otherwise in the opponent team. The parameter
	 *	<code>p</code> indicates the position of the player.
	 *	For example, the player's own position can be retrieved
	 *	with the method call 
	 *  <code>double[] s = player(true, getPosition())</code>.
	 *	Now, the player is in the coordinate position
	 *	(<code>s[X_COORDINATE]</code>, <code>s[Y_COORDINATE]</code>)
	 *	and has the heading <code>s[HEADING_ANGLE]</code> in radians.
	 *	
	 *	@see #X_COORDINATE
	 *	@see #Y_COORDINATE
	 *	@see #HEADING_ANGLE
	 */
	public double[] player(boolean us, int p);

	/**
	 *	Returns a team's messages. 
	 *	If <code>us == true</code>, the messages originate
	 *	from the own team, otherwise fromthe opponent team. 
	 *	For example, the message from the own team's
	 *	goalkeeper can be retrieved with the method call
	 *	<code>long msg = messages(true)[GOALKEEPER]</code>.
	 */
	public long[] messages(boolean us);
	
			
	/**
	 *	Returns the place of the faceoff.
	 *
	 *	@see #NO_FACEOFF
	 *	@see #CENTER_SPOT
	 *	@see #OUR_ENDZONE_LEFT
	 *	@see #OUR_ENDZONE_RIGHT
	 *	@see #OUR_NEUTRAL_ZONE_LEFT
	 *	@see #OUR_NEUTRAL_ZONE_RIGHT
	 *	@see #THEIR_NEUTRAL_ZONE_LEFT
	 *	@see #THEIR_NEUTRAL_ZONE_RIGHT
	 *	@see #THEIR_ENDZONE_LEFT
	 *	@see #THEIR_ENDZONE_RIGHT
	 */
	public int faceoff();

	/** 
	 *	Returns whether a team has scored a goal. 
	 *	If <code>us == true</code>, the team is own, 
	 *	otherwise the team is opponent.
	 */		
	public boolean isGoal(boolean us);
	
	/** 
	 *	Returns whether a team is in icing.
	 *	If <code>us == true</code>, the team is own, 
	 *	otherwise the team is opponent.
	 */		
	public boolean isIcing(boolean us);
	
	/** 
	 *	Returns whether a team is in offside. 
	 *	If <code>us == true</code>, the team is own, 
	 *	otherwise the team is opponent.
	 */		
	public boolean isOffside(boolean us);

	/** 
	 *	Returns whether any member of a team is interfering 
	 *  the other team's goalie. 
	 *	If <code>us == true</code>, the team is own, 
	 *	otherwise the team is opponent.
	 */		
	public boolean isInterferingGoalie(boolean us);

	/** 
	 *	Returns whether the goalkeeper of a team is
	 *	keeping the puck.
	 *	If <code>us == true</code>, the team is own, 
	 *	otherwise the team is opponent.
	 */		
	public boolean isGoalieKeepingPuck(boolean us);

	/** 
	 *	Returns whether a team has caused some other foul. 
	 *	If <code>us == true</code>, the team is own, 
	 *	otherwise the team is opponent.
	 */		
	public boolean isFoul(boolean us);
	
	/** 
	 *	Returns <code>true</code> if the puck has become
	 *	unplayable. 
	 */		
	public boolean isPuckUnplayable();
	
	}

