/* 
 * Server.java 1.0 2002-08-22
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.hockey.net;

import fi.utu.cs.hockey.*;
import fi.utu.cs.physics.*;
import fi.utu.cs.hockey.rink.FaceOff;

import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.UnknownHostException;
import java.net.SocketException;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.BufferedWriter;
import java.io.OutputStreamWriter;
import java.awt.Color;

/**
 *	A hockey game server.
 *
 *  @version 1.0, 020822
 *  @author  Jouni Smed
 */

public class Server //extends Thread //implements Runnable
	{
	/** Response when a client player is accepted to the home team. */
	static final String INIT_ACCEPTED_HOME			= "accepted to home";
	/** Response when a client player is accepted to the visiting team. */
	static final String INIT_ACCEPTED_VISITOR		= "accepted to visitor";
	/** 
	 *	Response when a client player is rejected because 
	 *  the position in a team is already taken. 
	 */
	static final String INIT_REJECTED_POSITION 	= "position already taken";
	/** 
	 *	Response when a client player is rejected because 
	 *  it does not belong to the home nor the visiting team.
	 */
	static final String INIT_REJECTED_TEAM 			= "already two teams";
	
	/** Message offset for the game state. */
	static final int REFEREE	= 0;
	/** Message offset for the game time. */
	static final int TIME  		= 1;
	/** Message offset for the game score. */
	static final int SCORE		= 2;
	/** Message offset for the puck's x-coordinate. */
	static final int PUCK_X		= 3;
	/** Message offset for the puck's y-coordinate. */
	static final int PUCK_Y		= 4;

	/** Message offset for a player's x-coordinate. */
	static final int OFFSET_X 			= 0;
	/** Message offset for a player's y-coordinate. */
	static final int OFFSET_Y 			= 1;
	/** Message offset for a player's orientation. */
	static final int OFFSET_ANGLE 	= 2;
	/** Message offset for a player's message. */
	static final int OFFSET_MSG 		= 3;
	/** The size of the player offset. */
	static final int OFFSET_SIZE		= 4;

	/** Message offset for the home team information. */
	static final int HOME			= 5;
	/** Message offset for the visiting team information. */
	static final int VISITOR 	= HOME + (6 * OFFSET_SIZE);		
	/** Total number of messages. */
	static final int MESSAGES	= VISITOR + (6 * OFFSET_SIZE);
	
	/** Referee bit indicating that the game is stopped. */
	static final long GAME_STOPPED										= 1L << 0;
	/** Referee bit indicating that the game is over. */	
	static final long GAME_OVER												=	1L << 1;
	
	/** Referee bit indicating a faceoff in the center spot. */
	static final long FACEOFF_CENTER 									= 1L << 2;
	/** Referee bit indicating a faceoff in the left home endzone spot. */
	static final long FACEOFF_HOME_ENDZONE_LEFT						= 1L << 3;
	/** Referee bit indicating a faceoff in the right home endzone spot. */
	static final long FACEOFF_HOME_ENDZONE_RIGHT					= 1L << 4;
	/** Referee bit indicating a faceoff in the left home neutral zone spot. */
	static final long FACEOFF_HOME_NEUTRAL_ZONE_LEFT			= 1L << 5;
	/** Referee bit indicating a faceoff in the right home neutral zone spot. */
	static final long FACEOFF_HOME_NEUTRAL_ZONE_RIGHT			= 1L << 6;
	/** Referee bit indicating a faceoff in the left visitor endzone spot. */
	static final long FACEOFF_VISITOR_ENDZONE_LEFT				= 1L << 7;
	/** Referee bit indicating a faceoff in the right visitor endzone spot. */
	static final long FACEOFF_VISITOR_ENDZONE_RIGHT				= 1L << 8;
	/** Referee bit indicating a faceoff in the left visitor neutral zone spot. */
	static final long FACEOFF_VISITOR_NEUTRAL_ZONE_LEFT		= 1L << 9;
	/** Referee bit indicating a faceoff in the right visitor neutral zone spot. */
	static final long FACEOFF_VISITOR_NEUTRAL_ZONE_RIGHT	= 1L << 10;
	
	/** Referee bit indicating a goal for the home team. */
	static final long HOME_GOAL 											= 1L << 11;
	/** Referee bit indicating an icing for the home team. */
	static final long HOME_ICING 											= 1L << 12;
	/** Referee bit indicating an offside for the home team. */
	static final long HOME_OFFSIDE 										= 1L << 13;
	/** Referee bit indicating the home team interfers the goalie. */
	static final long HOME_INTERFERING_GOALIE 					= 1L << 14;
	/** Referee bit indicating that the home team goalie has the puck. */
	static final long HOME_GOALIE_HAS_PUCK 						= 1L << 15;
	/** Referee bit indicating some other foul for the home team. */
	static final long HOME_OTHER 											= 1L << 16;
	
	/** Referee bit indicating a goal for the visiting team. */
	static final long VISITOR_GOAL 										= 1L << 17;
	/** Referee bit indicating an icing for the visiting team. */
	static final long VISITOR_ICING 									= 1L << 18;
	/** Referee bit indicating an offside for the visiting team. */
	static final long VISITOR_OFFSIDE					 				= 1L << 19;
	/** Referee bit indicating the visiting team interfers the goalie. */
	static final long VISITOR_INTERFERING_GOALIE 			= 1L << 20;
	/** Referee bit indicating that the visiting team goalie has the puck. */
	static final long VISITOR_GOALIE_HAS_PUCK 				= 1L << 21;
	/** Referee bit indicating some other foul for the visiting team. */
	static final long VISITOR_OTHER 									= 1L << 22;
	
	/** Referee bit indicating that the puck has become unplayable. */
	static final long PUCK_UNPLAYABLE									= 1L << 23;

	/** Mask for testing the authenticity. */
	static final long UNUSED_BITS	= 0xFFFFFFFFFFFFFFFFL << 24;

	/** Server port for commands. */
	private int serverPort;
	/** Server port for initialization. */
	private int initializationPort;
	/** Multicast group address. */	
	private InetAddress group;
	/** Multicast port. */
	private int groupPort;
	/** Socket for outgoing messages. */
	private DatagramSocket outputSocket;
	/** Packet for messages. */
	private DatagramPacket outputPacket;
	/** Socket for incoming commands. */
	private DatagramSocket inputSocket;
	/** Packet for commands. */
	private DatagramPacket inputPacket;

	/** Output for messages. */	
	private long[] output = new long[MESSAGES];
	/** Buffer for output packet. */
	private byte[] outputBuffer = new byte[MESSAGES * 8];
	/** Buffer for input packets. */
	private byte[] inputBuffer = new byte[Client.COMMANDS * 8];

	/** Direction of the attack. */
	private Vector2 attackDirection = new Vector2(0.0, 1.0);

	/** The game that is run in this server. */
	private Game game;

	/** Available positions in the home team. */
	private boolean[] homeFree = new boolean[Team.PLAYERS];	

	/** Available positions in the visiting team. */
	private boolean[] visitorFree = new boolean[Team.PLAYERS];	

	/**
	 *	Constructs a server for the hockey game <code>g</code>.
	 *	The server has two ports, one for initialization and
	 *	one for receiving input from the clients. The server sends
	 *	output to a multicast group. Hence, parameter <code>gAddr</code>
	 *	must be a D-class address (i.e., within the range
	 *  224.0.0.0-239.255.255.255). Preferably, it should be
	 *  from the range 239.0.0.0-239.255.255.255 which is intended
	 *  for local area network (LAN) use only.
	 *	
	 *	@param sPort Server port for client input
	 *	@param iPort Server port for client initialization
	 *	@param gAddr Multicast group for server output
	 *	@param gPort Multicast port
	 *	@param g		 Hockey game
	 *
	 */	
	public Server(int sPort, int iPort, String gAddr, int gPort, Game g)
		{
		serverPort = sPort;
		initializationPort = iPort;

		try
			{
			group = InetAddress.getByName(gAddr); 
			}
		catch (UnknownHostException e)
			{
			throw new RuntimeException("Unknown address " + gAddr +  ": " + e);
			}
		if (!group.isMulticastAddress())
			throw new IllegalArgumentException(gAddr + 
				" is not a multicast address.");
		groupPort = gPort;		
		
		try
			{
			outputSocket = new DatagramSocket();
			inputSocket = new DatagramSocket(serverPort);
			}
		catch (SocketException e)
			{
			throw new RuntimeException("Socket problem: " + e);
			}

		inputPacket = new DatagramPacket(inputBuffer, inputBuffer.length);
		outputPacket = new DatagramPacket(outputBuffer, outputBuffer.length, 
																			group, groupPort);
																			
		game = g;
		
		for (int i = 0; i < Team.PLAYERS; i++)
			{
			homeFree[i] = true;
			visitorFree[i] = true;
			}
		
		new InitializationHandler().start();
		new ServerHandler().start();
		}


	/** Sends the current game state to the clients. */
	public void sendGameState()
		{
			gameStateToOutput();
			outputToBuffer();
			try
				{
				outputSocket.send(outputPacket);
				}
			catch (IOException e)
				{
				throw new RuntimeException("Cannot send game state: " + e);
				}
		}


	/** Returns the referee message according to current game state. */
	private long refereeMessage()
		{
		if (game.isGameOver()) return GAME_OVER;
		
		long result = 0L;
		result |= (game.isGameStopped() ? GAME_STOPPED : 0L);
	
		if (game.isGameStopped())
			{
			int s = game.getFaceoffSpot();
			switch (s)
				{
				case FaceOff.CENTER_SPOT:
					result |= FACEOFF_CENTER;
					break;
				case FaceOff.HOME_ENDZONE_LEFT:
					result |= FACEOFF_HOME_ENDZONE_LEFT;
					break;
				case FaceOff.HOME_ENDZONE_RIGHT:
					result |= FACEOFF_HOME_ENDZONE_RIGHT;
					break;
				case FaceOff.HOME_NEUTRAL_ZONE_LEFT:
					result |= FACEOFF_HOME_NEUTRAL_ZONE_LEFT;
					break;
				case FaceOff.HOME_NEUTRAL_ZONE_RIGHT:
					result |= FACEOFF_HOME_NEUTRAL_ZONE_RIGHT;
					break;
				case FaceOff.VISITOR_ENDZONE_LEFT:
					result |= FACEOFF_VISITOR_ENDZONE_LEFT;
					break;
				case FaceOff.VISITOR_ENDZONE_RIGHT:
					result |= FACEOFF_VISITOR_ENDZONE_RIGHT;
					break;
				case FaceOff.VISITOR_NEUTRAL_ZONE_LEFT:
					result |= FACEOFF_VISITOR_NEUTRAL_ZONE_LEFT;
					break;
				case FaceOff.VISITOR_NEUTRAL_ZONE_RIGHT:
					result |= FACEOFF_VISITOR_NEUTRAL_ZONE_RIGHT;
					break;
				}
			}

		result |= (game.isGoal(true) ? HOME_GOAL : 0L);
		result |= (game.isIcing(true) ? HOME_ICING : 0L);
		result |= (game.isOffside(true) ? HOME_OFFSIDE : 0L);
		result |= (game.isInterferingGoalie(true) ? HOME_INTERFERING_GOALIE : 0L);
		result |= (game.isGoalieKeepingPuck(true) ? HOME_GOALIE_HAS_PUCK : 0L);
		result |= (game.isFoul(true) ? HOME_OTHER : 0L);

		result |= (game.isGoal(false) ? VISITOR_GOAL : 0L);
		result |= (game.isIcing(false) ? VISITOR_ICING : 0L);
		result |= (game.isOffside(false) ? VISITOR_OFFSIDE : 0L);
		result |= (game.isInterferingGoalie(false) ? VISITOR_INTERFERING_GOALIE : 0L);
		result |= (game.isGoalieKeepingPuck(false) ? VISITOR_GOALIE_HAS_PUCK : 0L);
		result |= (game.isFoul(false) ? VISITOR_OTHER : 0L);
	
		result |= (game.isPuckUnplayable() ? PUCK_UNPLAYABLE : 0L);
		
		return result;
		}


	/** Prepares the output messages according to the game state. */
	private void gameStateToOutput()
		{
		// General state information
		output[REFEREE] = refereeMessage();
		output[TIME] = game.getTime();
		output[SCORE] = ((long)game.getScore(true) << 32) 
									| (long)game.getScore(false);
		
		// Puck position
		output[PUCK_X] =
			Double.doubleToRawLongBits(game.getPuck().getPosition().getX());
		output[PUCK_Y] = 
			Double.doubleToRawLongBits(game.getPuck().getPosition().getY());
		
		// Home team positions, orientations, and messages
		int base = HOME;
		Team team =  game.getHome();		
		for (int i = 0; i < Team.PLAYERS; i++)
			{
			Player p = team.getPlayer(i);
			output[base + OFFSET_X] = 
				Double.doubleToRawLongBits(p.getPosition().getX());
			output[base + OFFSET_Y] = 
				Double.doubleToRawLongBits(p.getPosition().getY());
			double angle = p.getOrientation().angle(attackDirection);
			output[base + OFFSET_ANGLE] = Double.doubleToRawLongBits(angle);
			base += OFFSET_SIZE;
			}
			
		// Visiting team positions, orientations, and messages
		base = VISITOR;
		team =  game.getVisitor();
		for (int i = 0; i < Team.PLAYERS; i++)
			{
			Player p = team.getPlayer(i);
			output[base + OFFSET_X] = 
				Double.doubleToRawLongBits(p.getPosition().getX());
			output[base + OFFSET_Y] = 
				Double.doubleToRawLongBits(p.getPosition().getY());
			double angle = p.getOrientation().angle(attackDirection);
			output[base + OFFSET_ANGLE] = Double.doubleToRawLongBits(angle);
			base += OFFSET_SIZE;
			}
		}


	
	/** Copies the outputs to the packet buffer. */
	private void outputToBuffer()
		{
		for (int i = 0; i < output.length; i++)
			{
			long value = output[i];
			for (int j = ((i + 1) * 8) - 1; j >= (i * 8); j--)
				{
				outputBuffer[j] = (byte)value;
				value >>= 8;
				}
			}
		}


	/** 
	 *	Copies the packet buffer to the inputs, and sends the
	 *	commands to the game.
	 */
	private void bufferToGame()
		{
		long[] cmd = new long[Client.COMMANDS];

		// Extract the inputs
		for (int i = 0; i < cmd.length; i++)
			{
			long value = (long)inputBuffer[i * 8];
			value &= 0xFF;
			for (int j = ((i * 8) + 1); j < ((i + 1) * 8); j++)
				{
				value <<= 8;
				long nextByte = (long)inputBuffer[j];
				nextByte &= 0xFF;
				value |= nextByte;
				}
			cmd[i] = value;
			}
			
		// Decode and send the client's commands to the game
		int pos = 0;
		boolean home = false;
		if (cmd[Client.ID] > 0L)
			{
			pos = (int)cmd[Client.ID] - 1;
			home = true;
			}
		else
			{
			pos = (int)(-cmd[Client.ID]) - 1;
			home = false;
			}
		
		double shoot = Double.longBitsToDouble(cmd[Client.SHOOT]);
		double dash = Double.longBitsToDouble(cmd[Client.DASH]);
		double heading = Double.longBitsToDouble(cmd[Client.HEADING]);
		long say = cmd[Client.SAY];
		
		Team team = (home ? game.getHome() : game.getVisitor());
		Player player = team.getPlayer(pos);
		
		if (dash > 0.0) player.dash(dash);
		else if (dash < 0.0) player.brake(-dash);
		player.head(home ? heading : Math.PI + heading);
		if (player.canReach(game.getPuck()))
			{
			if (shoot > 0.0) game.shoot(player, shoot);
			else if (pos == Team.GOALKEEPER && shoot < 0.0) 
				game.keepPuck(player);
			}
		
		int base = (home ? HOME : VISITOR) + OFFSET_MSG;		
		output[base + (pos * OFFSET_SIZE)] = say;
		}



	/** Handles the command packets from the clients. */
	private class ServerHandler extends Thread
		{		
		/**	It's a daemon, which dies when the server dies. */
		public ServerHandler()
			{
			setDaemon(true);
			}

		/** Receives input packets and handles them. */
		public void run()
			{
			while (true)
				{
				try
					{
					inputSocket.receive(inputPacket);
					bufferToGame();
					}
				catch (IOException e)
					{
					throw new RuntimeException("Cannot receive input packet: " + e);
					}
				}
			}
		}



	/** Handles the client initialization call. */
	private class InitializationHandler extends Thread
		{
		/** A server socket to which initialization calls are directed. */
		private ServerSocket serverSocket;
		
		/** It's a daemon, which dies when the server dies. */
		public InitializationHandler()
			{
			setDaemon(true);
			try
				{
      	serverSocket = new ServerSocket(initializationPort);
      	}
      catch (IOException e)
      	{
      	throw new RuntimeException(
      		"Cannot open a server socket for initialization: " + e);
      	}
			}
		
		/** Receives the initialization call and handles it. */
	  public void run() 
	  	{
     	while (true) 
     		{
    		try 
    			{ 
        	Socket socket = serverSocket.accept();
        	initializePlayer(socket);
      		}
	    	catch (IOException e) 
    			{
					throw new RuntimeException(
						"Cannot open a socket for initialization: " + e);
	    		}
	    	}
    	}
    
		/**	Closes the server socket in the cleanup. */
    public void finalize() throws Throwable
    	{
    	try
    		{
    		if (serverSocket != null) serverSocket.close();
    		}
    	finally
    		{
    		super.finalize();
    		}
    	}
		}

	
	public boolean existsFreePosition()
		{
		for (int i = 0; i < Team.PLAYERS; i++)
			if (homeFree[i] | visitorFree[i]) return true;
		return false;
		}


	/** Handles client's initialization call from the socket <code>s</code>. */
	private void initializePlayer(Socket s) 
		{ 
		try
			{
			BufferedReader in = new BufferedReader(new
				InputStreamReader(s.getInputStream()));
			// Read player's name, number, position, team name, 
			// helmet color, and jersey color.
			String name = in.readLine();
			String number = in.readLine();
			String position = in.readLine();
			String teamName = in.readLine();
			String helmet = in.readLine();
			String jersey = in.readLine();

			PrintWriter out = new PrintWriter(new BufferedWriter(new
				OutputStreamWriter(s.getOutputStream())), true); 

			Color hc = new Color(Integer.parseInt(helmet), true);
			Color jc = new Color(Integer.parseInt(jersey), true);

			if (game.getHome().getName() == null
				|| game.getHome().getName().equals(teamName))
				{
				int pos = Integer.parseInt(position);
				
				if (!homeFree[pos])
					{
					// Player is rejected because position is already taken.
					out.println(INIT_REJECTED_POSITION);
					}
				else
					{
					// Player is accepted to the home team
					homeFree[pos] = false;
					Team team = game.getHome();
					team.setName(teamName);
					Player p = team.getPlayer(pos);
					p.setName(name);
					p.setNumber(Integer.parseInt(number));
					p.setColors(hc, jc);
					out.println(INIT_ACCEPTED_HOME);
					game.teamUpdated();
					}
				}
			else if (game.getVisitor().getName() == null 
				|| game.getVisitor().getName().equals(teamName))
				{
				int pos = Integer.parseInt(position);
				
				if (!visitorFree[pos])
					{
					// Player is rejected because position is already taken.
					out.println(INIT_REJECTED_POSITION);
					}
				else
					{
					// Player is accepted to the visiting team
					visitorFree[pos] = false;
					Team team = game.getVisitor();
					team.setName(teamName);
					Player p = team.getPlayer(pos);
					p.setName(name);
					p.setNumber(Integer.parseInt(number));
					p.setColors(hc, jc);
					out.println(INIT_ACCEPTED_VISITOR);
					game.teamUpdated();
					}
				}
			else
				{
				// Player is rejected because there are two teams already
				out.println(INIT_REJECTED_TEAM);
				}
			}
		catch (IOException e) 
			{
			throw new RuntimeException("Player initialization failed: " + e);
			} 
		finally 
			{
			try 
				{
				s.close();
				} 
			catch (IOException e) 
				{
				throw new RuntimeException("Cannot close socket: " + e);
				} 
			}
		}
		
	}