/* 
 * Rink.java 1.0 2002-06-05
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
package fi.utu.cs.hockey.rink;

import fi.utu.cs.physics.*;
import java.awt.Color;

/**
 *  Definitions for an ice hockey rink. All measures are given
 *	in meters, and they follow the
 *  <a href="http://www.iihf.com/hockey/rules/offrules.htm">IIHF 
 *  Official Rules</a>.
 *
 *  @version 1.0, 020605
 *  @author  Jouni Smed
 */
public class Rink 
	{
	/** Width of the ice hockey rink. */
	public static final double WIDTH = 30.00;
	/** Length of the ice hockey rink. */
	public static final double LENGTH = 60.00;
	/** Radius of the corner boards. */
	public static final double BOARD_RADIUS = 8.00;
	/** Width of the corner boards. */
	public static final double BOARD_WIDTH = 0.10;
	/** Width of the goal line. */
	public static final double GOAL_LINE_WIDTH = 0.05;
	/** Width of the blue line. */
	public static final double BLUE_LINE_WIDTH = 0.30;
	/** Width of the center line. */
	public static final double CENTER_LINE_WIDTH = 0.30;
	/** Distance from the center line to the goal line. */
	public static final double GOAL_LINE_OFFSET = 26.00;
	/** The length of the goal line between the rounded corners. */
	public static final double GOAL_LINE_LENGTH = 27.86;
	/** Distance from the center line to the blue line. */
	public static final double BLUE_LINE_OFFSET =	8.60;
	/** Radius of the center circle. */
	public static final double CENTER_CIRCLE_RADIUS =	4.50;	
	/** Width of the center circle. */
	public static final double CENTER_CIRCLE_WIDTH = 0.05;	
	/** Radius of the center spot. */
	public static final double CENTER_ICE_SPOT_RADIUS = 0.15;  
	/** Radius of the face off circle. */
	public static final double FACE_OFF_CIRCLE_RADIUS	= 4.50;	
	/** Width of the face off circle. */
	public static final double FACE_OFF_CIRCLE_WIDTH = 0.05;
	/** Radius of the face off spot. */
	public static final double FACE_OFF_SPOT_RADIUS =	0.30;  
	/** Distance from the center to the face off spot. */
	public static final double FACE_OFF_OFFSET = 7.00; 
	/** Distance from the center line to the endzone face off spot. */
	public static final double ENDZONE_FACE_OFF_OFFSET = 20.00; 
	/** Distance from the center line to the neutral zone face off spot. */
	public static final double NEUTRAL_ZONE_FACE_OFF_OFFSET	= 7.10;
	/** Radius of the goal crease. */	
	public static final double GOAL_CREASE_RADIUS =	1.80;
	/** Width of the goal crease. */	
	public static final double GOAL_CREASE_WIDTH = 0.05;
	/** Width of the goal. */		
	public static final double GOAL_WIDTH = 1.83;
	/** Width of the goal post. */	
	public static final double GOAL_POST_WIDTH = 0.05;
	/** Distance from the center line to the scoring line. */
	public static final double SCORE_LINE_OFFSET = GOAL_LINE_OFFSET + 
																					(GOAL_LINE_WIDTH / 2.0);
	/** Width of the goal net. */		
	public static final double GOAL_NET_WIDTH = 1.59;
	/** Distance from the center line to the goal net. */
	public static final double GOAL_NET_OFFSET = GOAL_LINE_OFFSET + 0.45;
	
	/** The color used in the ice. */	
	public static final Color COLOR_ICE = Color.white;
	/** The color used in the boards. */	
	public static final Color COLOR_BOARD = Color.black;
	/** The red color used in the center line, goal lines, face off cirles,
	    face off spots, goal crease lines, and goal posts. */	
	public static final Color	COLOR_RED = Color.red;
	/** The blue color used in the center ice spot, center circle, and
	    blue lines. */
	public static final Color COLOR_BLUE = Color.blue;
	/** The light blue color used in the goal crease area. */
	public static final Color COLOR_LIGHT_BLUE = new Color(0xC0, 0xC0, 0xFF);


	static final boolean HALF_A = true;
	static final boolean HALF_B = false;
	static final boolean SIDE_A = true;
	static final boolean SIDE_B = false;
	static final boolean ENDZONE = true;
	static final boolean NEUTRAL_ZONE = false;
	
	private CenterIceSpot centerIceSpot;
	private CenterCircle centerCircle;
	private FaceOffSpot[] faceOffSpots;
	private FaceOffCircle[] faceOffCircles;
	private CenterLine centerLine;
	private BlueLine[] blueLines;
	private GoalLine[] goalLines;
	private GoalPost[] goalPosts;
	private Goal[] goals;
	private GoalCrease[] goalCreases;
	private SideBoard[] sideBoards;
	private EndBoard[] endBoards;
	private CornerBoard[] cornerBoards;
	private GoalNet[] goalNets;
	public ScoreLine[] scoreLines;           // <----- !!!!!!!!!!!!!
	
	/** Creates a rink in the universe <code>u</code>. */
	public Rink(Universe u)
		{
		centerIceSpot = new CenterIceSpot();
		centerCircle = new CenterCircle();
		
		faceOffSpots = new FaceOffSpot[8];
		faceOffSpots[index(HALF_A, SIDE_A, ENDZONE)]
						    = new FaceOffSpot(HALF_A, SIDE_A, ENDZONE);
		faceOffSpots[index(HALF_A, SIDE_A, NEUTRAL_ZONE)]
							  = new FaceOffSpot(HALF_A, SIDE_A, NEUTRAL_ZONE);
		faceOffSpots[index(HALF_A, SIDE_B, ENDZONE)]
			 				  = new FaceOffSpot(HALF_A, SIDE_B, ENDZONE);
		faceOffSpots[index(HALF_A, SIDE_B, NEUTRAL_ZONE)]
								= new FaceOffSpot(HALF_A, SIDE_B, NEUTRAL_ZONE);
		faceOffSpots[index(HALF_B, SIDE_A, ENDZONE)]
						    = new FaceOffSpot(HALF_B, SIDE_A, ENDZONE);
		faceOffSpots[index(HALF_B, SIDE_A, NEUTRAL_ZONE)]
							  = new FaceOffSpot(HALF_B, SIDE_A, NEUTRAL_ZONE);
		faceOffSpots[index(HALF_B, SIDE_B, ENDZONE)]
			 				  = new FaceOffSpot(HALF_B, SIDE_B, ENDZONE);
		faceOffSpots[index(HALF_B, SIDE_B, NEUTRAL_ZONE)]
								= new FaceOffSpot(HALF_B, SIDE_B, NEUTRAL_ZONE);

		faceOffCircles = new FaceOffCircle[4];		
		faceOffCircles[index(HALF_A, SIDE_A)]
									= new FaceOffCircle(HALF_A, SIDE_A);
		faceOffCircles[index(HALF_A, SIDE_B)]
									= new FaceOffCircle(HALF_A, SIDE_B);
		faceOffCircles[index(HALF_B, SIDE_A)]
									= new FaceOffCircle(HALF_B, SIDE_A);
		faceOffCircles[index(HALF_B, SIDE_B)]
									= new FaceOffCircle(HALF_B, SIDE_B);

		centerLine = new CenterLine();
		
		blueLines = new BlueLine[2];
		blueLines[index(HALF_A)] = new BlueLine(HALF_A);
		blueLines[index(HALF_B)] = new BlueLine(HALF_B);

		goalLines = new GoalLine[2];
		goalLines[index(HALF_A)] = new GoalLine(HALF_A);
		goalLines[index(HALF_B)] = new GoalLine(HALF_B);

		goalPosts = new GoalPost[4];		
		goalPosts[index(HALF_A, SIDE_A)]
									= new GoalPost(HALF_A, SIDE_A);
		goalPosts[index(HALF_A, SIDE_B)]
									= new GoalPost(HALF_A, SIDE_B);
		goalPosts[index(HALF_B, SIDE_A)]
									= new GoalPost(HALF_B, SIDE_A);
		goalPosts[index(HALF_B, SIDE_B)]
									= new GoalPost(HALF_B, SIDE_B);
	
		goals = new Goal[2];
		goals[index(HALF_A)] = new Goal(HALF_A);
		goals[index(HALF_B)] = new Goal(HALF_B);

		goalCreases = new GoalCrease[2];
		goalCreases[index(HALF_A)] = new GoalCrease(HALF_A);
		goalCreases[index(HALF_B)] = new GoalCrease(HALF_B);

		sideBoards = new SideBoard[2];
		sideBoards[index(SIDE_A)] = new SideBoard(SIDE_A);
		sideBoards[index(SIDE_B)] = new SideBoard(SIDE_B);

		endBoards = new EndBoard[2];
		endBoards[index(HALF_A)] = new EndBoard(HALF_A);
		endBoards[index(HALF_B)] = new EndBoard(HALF_B);
		
		cornerBoards = new CornerBoard[4];		
		cornerBoards[index(HALF_A, SIDE_A)]
									= new CornerBoard(HALF_A, SIDE_A);
		cornerBoards[index(HALF_A, SIDE_B)]
									= new CornerBoard(HALF_A, SIDE_B);
		cornerBoards[index(HALF_B, SIDE_A)]
									= new CornerBoard(HALF_B, SIDE_A);
		cornerBoards[index(HALF_B, SIDE_B)]
									= new CornerBoard(HALF_B, SIDE_B);

		goalNets = new GoalNet[2];
		goalNets[index(HALF_A)] = new GoalNet(HALF_A);
		goalNets[index(HALF_B)] = new GoalNet(HALF_B);
		
		scoreLines = new ScoreLine[2];
		scoreLines[index(HALF_A)] = new ScoreLine(HALF_A);
		scoreLines[index(HALF_B)] = new ScoreLine(HALF_B);
		
		
		u.addBoundary(centerIceSpot);
		u.addBoundary(centerCircle);
		for (int i = 0; i < 8; i++) 
			u.addBoundary(faceOffSpots[i]);
		for (int i = 0; i < 4; i++) 
			u.addBoundary(faceOffCircles[i]);
		u.addBoundary(centerLine);
		for (int i = 0; i < 2; i++) 
			u.addBoundary(blueLines[i]);
		for (int i = 0; i < 2; i++) 
			u.addBoundary(goalLines[i]);
		for (int i = 0; i < 2; i++) 
			u.addBoundary(goalNets[i]);
		for (int i = 0; i < 2; i++) 
			u.addBoundary(scoreLines[i]);
		for (int i = 0; i < 2; i++) 
			u.addBoundary(goals[i]);
		for (int i = 0; i < 4; i++)
			u.addBoundary(goalPosts[i]);
		for (int i = 0; i < 2; i++) 
			u.addBoundary(goalCreases[i]);
		for (int i = 0; i < 2; i++) 
			u.addBoundary(sideBoards[i]);
		for (int i = 0; i < 2; i++) 
			u.addBoundary(endBoards[i]);
		for (int i = 0; i < 4; i++) 
			u.addBoundary(cornerBoards[i]);
		u.addBoundary(new Ice());
		}

	private int index(boolean half, boolean side, boolean zone)
		{
		int result = 0;
		if (half) result |= 1;
		if (side) result |= 2;
		if (zone) result |= 4;
		return result;
		}

	private int index(boolean half, boolean side)
		{
		int result = 0;
		if (half) result |= 1;
		if (side) result |= 2;
		return result;
		}
	
	private int index(boolean half)
		{
		return half ? 1 : 0;
		}
	}

