/* 
 * Boundary.java 1.1 2002-06-11
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

/**
 *  A static physical boundary which can be open or closed.  
 *  The boundary can be owned by an instance of the class 
 *  <code>Universe</code>. The boundary can be either corporeal or 
 *  ethereal, which determines whether it can be collided or not 
 *  (i.e., only the corporeal boundaries can be collided).
 * 
 *  @see Universe
 *  @see Body
 *  @since 1.1
 *  @version 1.1, 020611
 *  @author  Jouni Smed
 */

public abstract class Boundary extends Particle
  {      
	/** Maximum accepted round-off difference in comparisons. */
	private final static double INITIAL_BOUNDARY_EPSILON = 0.0000000001;
	/** An epsilon comparator for boundary calculations. */
	protected final static EpsilonComparator boundaryEpsilon = 
		new EpsilonComparator(INITIAL_BOUNDARY_EPSILON);

  /** 
   *  Constructs a boundary that has no owner.  
   */
  protected Boundary()
    {
    super();
    }


	/**
	 *	Returns <code>true</code>, if the bounding box of the path 
	 *  (<code>x1</code>, <code>y1</code>)-(<code>x2</code>, <code>y2</code>)
	 *	of a circle with a radius <code>r</code> crosses the boundary; 
	 *  otherwise, returns <code>false</code>.
	 */
	protected abstract boolean isCrossedByBoundingBox(double x1, double y1,
																			double x2, double y2, double r);


	/**
	 *	Returns <code>true</code>, if the line (<code>x1</code>,
	 *  <code>y1</code>)-(<code>x2</code>, <code>y2</code>)
	 *	crosses the boundary; otherwise, returns <code>false</code>.
	 */
	protected abstract boolean isCrossedByLine(double x1, double y1,
					 																	 double x2, double y2);

  /** 
   *  Checks whether the boundary will be crossed by the body 
   *  <code>other</code> within the next <code>dt</code> seconds. 
   *  @see Body#advance(double)
   *  @see #isCrossedByBoundingBox(double,double,double,double,double)
   *  @see #isCrossedByLine(double,double,double,double)
   */
  public boolean isCrossed(Body other, double dt)
    {
    Point2 otherCurrentPosition = other.getPosition();
		Point2 otherFuturePosition = other.futurePosition(dt);

  	double x1 = otherCurrentPosition.getX();
  	double y1 = otherCurrentPosition.getY();
  	
  	double x2 = otherFuturePosition.getX();
  	double y2 = otherFuturePosition.getY();
  	
  	double r = other.getRadius();

		// Is there any movement at all?
		if (x1 == x2 & y1 == y2) return false;

		// Does the bounding box cross the boundary?
		if (!isCrossedByBoundingBox(x1, y1, x2, y2, r)) return false;

		// Calculate three points around the future position:
		//   (i) future position + radius along the tangent
		//  (ii) future position + radius along the normal
		// (iii) future position - radius along the normal
		double deltaX = x2 - x1;
		double deltaY = y2 - y1;

		double q = r / Math.sqrt(deltaX * deltaX + deltaY * deltaY);
		double dx = q * deltaX;
		double dy = q * deltaY;

		double rx1 = 0.0;
		double ry1 = 0.0;
		double rx2 = 0.0;
		double ry2 = 0.0;
		
		if (boundaryEpsilon.compare(deltaX, 0.0) == 0)
			{
			rx1 = 0.0;
			ry1 = r;
			rx2 = 0.0;
			ry2	= r;
			}
		else if (boundaryEpsilon.compare(deltaY, 0.0) == 0)
			{
			rx1 = r;
			ry1 = 0.0;
			rx2 = -r;
			ry2	= 0.0;
			}
		else
			{
			double alpha = Math.atan(deltaY / deltaX);
			double beta = alpha + Math.PI / 2.0;
			double gamma = alpha - Math.PI / 2.0;
			rx1 = Math.sin(beta) * r;
			ry1 = Math.cos(beta) * r;
			rx2 = Math.sin(gamma) * r;
			ry2 = Math.cos(gamma) * r;
			}

		// Test whether any of the line from current position to
		// the three calculated positions crosses the boundary.
		return isCrossedByLine(x1, y1, x2 + dx, y2 + dy) 
				|| isCrossedByLine(x1, y1, x2 + rx1, y2 + ry1)
				|| isCrossedByLine(x1, y1, x2 + rx2, y2 + ry2);
    }
  
  }