/* 
 * BoundaryCurve.java 1.1 2002-06-11
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

/**
 *  A boundary curve. 
 * 
 *  @since 1.1
 *  @version 1.1, 020611
 *  @author  Jouni Smed
 */

public class BoundaryCurve extends Boundary
  {    
  /** Center point of the boundary curve. */
  private Point2 center;
  /** Radius of the boundary curve. */
  private double radius;
  /** Start angle in radians. */
  private double startAngle;
  /** Angular extent in radians. */
  private double angularExtent;
	/** Angle of the last detected collision. */
	private double lastCollisionAngle;

  /** 
   *  Constructs a boundary circle with a center in point (<code>px</code>, 
   *  <code>py</code>) and a radius <code>r</code>
   */
  public BoundaryCurve(double px, double py, double r)
  	{
  	this(new Point2(px, py), r, 0.0, 2.0 * Math.PI);
  	}

  /** 
   *  Constructs a boundary circle with a center in point <code>p</code> 
   *  and a radius <code>r</code>
   */
  public BoundaryCurve(Point2 p, double r)
  	{
  	this(p, r, 0.0, 2.0 * Math.PI);
  	}
  
  /** 
   *  Constructs a boundary curve with a center in point (<code>px</code>, 
   *  <code>py</code>), a radius <code>r</code>, a start angle <code>a</code>,
   *  and an angular extent <code>e</code>. The angles are expressed 
   *  in radians clockwise. 
   */
  public BoundaryCurve(double px, double py, double r, double a, double e)
  	{
  	this(new Point2(px, py), r, a, e);
  	}
  
  
  /** 
   *  Constructs a boundary curve with a center in point <code>p</code>,
   *  a radius <code>r</code>, a start angle <code>a</code>,
   *  and an angular extent <code>e</code>. The angles are expressed 
   *  in radians clockwise. 
   */
  public BoundaryCurve(Point2 p, double r, double a, double e)
    {
    super();
    center = (Point2)p.clone();
    radius = r;
    startAngle = a;
    angularExtent = e;
    }
  
  /** Returns the center of the boundary curve. */
  public Point2 getCenter()
  	{
  	return (Point2)center.clone();
  	}
  
  /** Returns the radius of the boundary curve. */
  public double getRadius()
  	{
  	return radius;
  	}

  /** Returns the start angle of the boundary curve. */
  public double getStartAngle()
  	{
  	return startAngle;
  	}

  /** Returns the angular extent of the boundary curve. */
  public double getAngularExtent()
  	{
  	return angularExtent;
  	}
  

	/**
	 *	Returns <code>true</code>, if the bounding box of the path 
	 *  (<code>x1</code>, <code>y1</code>)-(<code>x2</code>, <code>y2</code>)
	 *	of a circle with a radius <code>r</code> crosses the boundary curve; 
	 *  otherwise, returns <code>false</code>.
	 */
	protected boolean isCrossedByBoundingBox(double x1, double y1,
				 																	 double x2, double y2, double r)
		{
		double xmin = Math.min(x1 - r, x2 - r);
		double ymin = Math.min(y1 - r, y2 - r);
		double xmax = Math.max(x1 + r, x2 + r);
		double ymax = Math.max(y1 + r, y2 + r);
  	double xc = center.getX();
  	double yc = center.getY();
		
		return !(xmin > xc + radius | xmax < xc - radius | 
		 			   ymin > yc + radius | ymax < yc - radius);
		}


	/**
	 *	Returns <code>true</code>, if the line (<code>x1</code>,
	 *  <code>y1</code>)-(<code>x2</code>, <code>y2</code>)
	 *	crosses the boundary curve; otherwise, returns <code>false</code>.
	 */
	protected boolean isCrossedByLine(double x1, double y1, double x2, double y2)
		{
  	double xc = center.getX();
  	double yc = center.getY();
  	
  	double r = radius; 

		double deltaX = x2 - x1;
		double deltaY = y2 - y1;

  	double a = (deltaX * deltaX) + (deltaY * deltaY);
  	double b = 2 * ((deltaX * (x1 - xc)) + (deltaY * (y1 - yc)));
  	double c = (xc * xc) + (yc * yc) + (x1 * x1) + (y1 * y1) 
  							- 2 * ((xc * x1) + (yc * y1)) - (r * r);

  	// Does the line intersect at all?
  	double t = (b * b) - (4.0 * a * c);
  	if (t < 0.0) return false;
 	
  	double u = ((xc - x1) * (x2 - x1) + (yc - y1) * (y2 - y1)) /
  						 ((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));

		double xp = x1 + u * (x2 - x1);
		double yp = y1 + u * (y2 - y1);
		// Is the closest point within the circle?
		if (((xp - xc) * (xp - xc)) + ((yp - yc) * (yp - yc)) > (r * r))	
			return false;
		
		t = Math.sqrt(t);
		double u1 = (-b + t) / (2 * a);
		double u2 = (-b - t) / (2 * a);
		// Does the line segment cross the circle?
		if ((u1 < 0.0 | u1 > 1.0) 
			& (u2 < 0.0 | u2 > 1.0)) return false;

		if (u2 >= 0.0 & u2 <= 1.0)
			{
			double x = x1 + u2 * (x2 - x1);
			double y = y1 + u2 * (y2 - y1);
			Vector2 v2 = new Vector2(x - xc, y - yc);
			Vector2 v0 = new Vector2(1.0, 0.0);
			double phi = v0.angle(v2);
			Vector2 vn = new Vector2(0.0, 1.0);
			if (vn.project(v2).getDeltaY() > 0.0) phi = (2.0 * Math.PI - phi);
			// Is the intersecting point within the curve?
			lastCollisionAngle = phi;			
			if (startAngle <= phi & phi <= (startAngle + angularExtent)) 
				return true;
			}
		if (u1 >= 0.0 & u1 <= 1.0)
			{
			double x = x1 + u1 * (x2 - x1);
			double y = y1 + u1 * (y2 - y1);
			Vector2 v1 = new Vector2(x - xc, y - yc);
			Vector2 v0 = new Vector2(1.0, 0.0);
			double psi = v0.angle(v1);
			Vector2 vn = new Vector2(0.0, 1.0);
			if (vn.project(v1).getDeltaY() > 0.0) psi = (2.0 * Math.PI - psi);
			// Is the intersecting point within the curve?
			lastCollisionAngle = psi;			
			if (startAngle <= psi & psi <= (startAngle + angularExtent)) 
				return true;
			}
		// If all else fails, it does not intersect.
		return false;
		}


  /**
   *  Realizes an elastic collision for the static boundary and
   *  the body <code>other</code>.
   *  The linear momentum and kinetic energy of the body is conserved 
   *  in the collision.
   */
  public void elasticCollision(Body other)
    {    
    Vector2 v2 = other.getVelocity();
    Vector2 a2 = other.getAcceleration();

		Vector2 vn = new Vector2(1.0, 0.0);
		vn.rotate(-lastCollisionAngle);
		Vector2 vt = (Vector2)vn.clone();
		vt.rotate90();
		
		Vector2 vn2 = vn.project(v2);
		Vector2 vt2 = vt.project(v2);		
		Vector2 an2 = vn.project(a2);
		Vector2 at2 = vt.project(a2);

		vn2.rotate180();
		an2.rotate180();
		
		vn2.add(vt2);
		an2.add(at2);

		other.setVelocity(vn2);		
    other.setAcceleration(an2);
		}


  /**
   *  Realizes an inelastic collision for the static boundary and
   *  the body <code>other</code>.
   *  The linear momentum is conserved but kinetic energy is lost 
   *  in the collision.
   */
  public void inelasticCollision(Body other)
    {    
		other.setVelocity(0.0, 0.0);		
    other.setAcceleration(0.0, 0.0);
		}


  /** Returns a string that contains the values of the attributes. */
  public String toString()
    {
    return "center = " + center + " radius = " + radius + " startAngle = " 
    	+ startAngle + " angularExtent = " + angularExtent + " " 
    	+ super.toString();
    }
	}
