/* 
 * EpsilonComparator.java 1.0 2001-05-08
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

import java.util.Comparator;

/**
 *	A rounding comparator for double-precision floating point values.
 * 	An epsilon value is used when comparing two values 
 *  (i.e., <code>compare(f, g)</code> returns zero, if 
 *  <code>Math.abs(f - g) < getComparisonEpsilon()</code>).
 *
 *	@since 1.1
 *  @version 1.1, 020508
 *  @author  Jouni Smed
 */


public class EpsilonComparator implements Comparator
	{
	/** The value of epsilon. */
	private double epsilon;
	
	/** Creates a comparator with the epsilon value <code>e</code>. */
	public EpsilonComparator(double e)
		{
		setEpsilon(e);
		}
	
	/** Returns the epsilon. */
	public double getEpsilon()
		{
		return epsilon;
		}
	
	/**
	 *	Sets the epsilon value to <code>e</code>. The new
	 *  epsilon must be greater than zero.
	 */
	public void setEpsilon(double e)
		{
		if (e <= 0.0) throw new IllegalArgumentException(
			"Epsilon must be greater than zero.");
		epsilon = e;
		}
	
	/** 
	 * 	Compares <code>v1</code> and <code>v2</code> for order. 
	 *  Returns a negative integer, zero, or a positive integer as 
	 *  <code>v1</code> is less than, equal to, or greater than 
	 *  <code>v2</code>. The values are equal, if their difference
	 *  is less than epsilon. 
	 *
	 */
	public int compare(double v1, double v2)
		{
		if (Math.abs(v1 - v2) < epsilon) return 0;
		return (v1 < v2 ?  -1 : 1);
		}
	
	/** 
	 * 	Compares the two arguments for order. The arguments must 
	 *  be inherited from the class <code>Number</code>. 
	 *  Returns a negative integer, zero, or a positive integer as the 
	 *  first argument is less than, equal to, or greater than the second. 
	 *  The arguments are equal, if their difference is less than epsilon.
	 *	Note: This comparator imposes orderings that are inconsistent
	 *	with equals.
	 */
	public int compare(Object o1, Object o2)
		{
		Number d1 = (Number)o1;
		Number d2 = (Number)o2;
		return compare(d1.doubleValue(), d2.doubleValue());
		}
	
	}
