/* 
 * FrictionalBody.java 1.1 2002-06-04
 *
 *  2001-2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

import java.util.Set;
import java.util.HashSet;

/**
 *  A moving physical body with mass and dimensionality. The outline of 
 *  the body is approximated with a bounding circle of a given radius, 
 *  which is used in detecting overlaps. The body can be owned by
 *  an instance of the class <code>Universe</code>, which may impose limits
 *  to the position. 
 * 
 *  @since 1.1
 *  @version 1.1, 020604
 *  @author  Jouni Smed
 */

public abstract class FrictionalBody extends Body 
  {    
  /** Static and kinetic friction of the body. */
	private Friction friction;
	/** Maximum accepted round-off difference in comparisons. */
	private final static double INITIAL_FRICTION_EPSILON = 0.1;
	/** An epsilon comparator for friction calculations. */
	private final static EpsilonComparator frictionEpsilon = 
		new EpsilonComparator(INITIAL_FRICTION_EPSILON);
  
  /** 
   *  Constructs a body with the mass <code>m</code>, 
   *  the radius <code>r</code> and friction <code>f</code>. 
   *  The body resides in the 
   *  origin and has a zero velocity and a zero acceleration.
   *  The body is corporeal and has no owner.  
   */
  protected FrictionalBody(double m, double r, Friction f)
    {
    this(m, r, Point2.ORIGIN, Vector2.ZERO, Vector2.ZERO, f);
    }

  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code>, the position <code>p</code>
   *  and friction <code>f</code>. 
   *  The body has a zero velocity and a zero acceleration.
   *  The body is corporeal and has no owner.  
   */
  protected FrictionalBody(double m, double r, Point2 p, Friction f)
    {
    this(m, r, p, Vector2.ZERO, Vector2.ZERO, f);
    }
  
  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code>, the position <code>p</code>,
   *  the velocity <code>v</code> and friction <code>f</code>.  
   *  The body has a 
   *  zero acceleration.   
   *  The body is corporeal and has no owner.  
   */
  protected FrictionalBody(double m, double r, Point2 p, Vector2 v, Friction f)
    {
    this(m, r, p, v, Vector2.ZERO, f);
    }

  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code>, the position <code>p</code>,
   *  the velocity <code>v</code>, the acceleration <code>a</code>
   *  and friction <code>f</code>.
   *  The body is corporeal and has no owner.  
   */
  protected FrictionalBody(double m, double r, Point2 p, Vector2 v, 
  							 Vector2 a, Friction f)
    {
    super(m, r, p, v, a);
    friction		 = (Friction)f.clone();
    setOwner(null);
    }

  /** Returns the friction of the body. */
  public Friction getFriction()
    {
    return (Friction)friction.clone();
    }

  /** Sets the friction of the body to <code>f</code>. */
  public void setFriction(Friction f)
    {
    friction = (Friction)f.clone();
    }
  
	/**
	 *	Returns the velocity of the body after <time>dt</time> seconds.
	 *  If the body is not frictionless, friction slows down the velocity.
	 */	 
	public Vector2 futureVelocity(double dt)
		{
    double g = (getOwner() != null ? 
    						getOwner().getAccelerationOfGravity() : 0.0);

		Vector2 movement = new Vector2(getPosition(), futurePosition(dt));
		Vector2 a = getAcceleration();
		Vector2 v = getVelocity();
		
		boolean frictionIsKinetic =
			frictionEpsilon.compare(movement.length(), 0.0) != 0;
		
		if (!frictionIsKinetic && 
				(friction.getStaticCoefficient() * g) > getAcceleration().length())
			{
			a.set(0.0, 0.0);
			v.set(0.0, 0.0);
			}
		else
			{			
			Vector2 at = movement.project(a);
			movement.rotate90();
			a = movement.project(a);

			movement.rotate90();			
			Vector2 af = movement.unitVector();
			af.multiply(friction.getKineticCoefficient() * g);
			at.add(af);
		  a.add(at);
			}

    double vx = v.getDeltaX();
    double vy = v.getDeltaY();
    double ax = a.getDeltaX();
    double ay = a.getDeltaY();
    vx = (ax * dt) + vx;
    vy = (ay * dt) + vy;
    return new Vector2(vx, vy);
		}		


  /** Returns a string that contains the values of the attributes. */
  public String toString()
    {
    return "friction = " + friction + " " + super.toString();
    }
  }