/* 
 * OrientedBody.java 1.1 2002-05-08
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

/**
 *  A moving physical body with mass, dimensionality and orientation. 
 *  The body can have static and kinetic friction.
 *
 *	@since 1.1
 *  @version 1.1, 020508
 *  @author  Jouni Smed
 *	@see Friction
 */

public abstract class OrientedBody extends Body 
  {
  /** The orientation of this object. */
  private Vector2 orientation;
  /** Coefficients for static and kinetic tangent friction. */
  private Friction tangentFriction;
  /** Coefficients for static and kinetic normal friction. */
  private Friction normalFriction;
  /** The type of current tangent friction. */
	private boolean tangentFrictionIsKinetic;
	/** The type of current normal friction. */
	private boolean normalFrictionIsKinetic;
	/** Maximum accepted round-off difference in comparisons. */
	private final static double INITIAL_FRICTION_EPSILON = 0.1;
	/** An epsilon comparator for friction calculations. */
	private final static EpsilonComparator frictionEpsilon = 
		new EpsilonComparator(INITIAL_FRICTION_EPSILON);

  /** 
   *  Constructs a body with the mass <code>m</code>, 
   *  the radius <code>r</code> and the orientation <code>o</code>. 
   *  The body resides in the origin and has a zero velocity and 
   *  a zero acceleration. The body is corporeal and frictionless 
   *  and has no owner.  
   */
  protected OrientedBody(double m, double r, Vector2 o)
    {
    this(m, r, o, Point2.ORIGIN, Vector2.ZERO, Vector2.ZERO);
    }

  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code>, the orientation <code>o</code> 
   *  and the position <code>p</code>. 
   *  The body has a zero velocity and a zero acceleration.
   *  The body is corporeal and frictionless and has no owner.  
   */
  protected OrientedBody(double m, double r, Vector2 o, Point2 p)
    {
    this(m, r, o, p, Vector2.ZERO, Vector2.ZERO);
    }
  
  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code>, the orientation <code>o</code>,
   *  the position <code>p</code>
   *  and the velocity <code>v</code>.  The body has a 
   *  zero acceleration. The body is corporeal and frictionless and 
   *  has no owner.  
   */
  protected OrientedBody(double m, double r, Vector2 o, Point2 p, Vector2 v)
    {
    this(m, r, o, p, v, Vector2.ZERO);
    }

  /** 
   *  Constructs a body with the mass <code>m</code>,
   *  the radius <code>r</code>, the orientation <code>o</code>,
   *  the position <code>p</code>,
   *  the velocity <code>v</code> and the acceleration <code>a</code>.
   *  The body is corporeal and frictionless and has no owner.  
   */
  protected OrientedBody(double m, double r, Vector2 o, 
                          Point2 p, Vector2 v, Vector2 a)
    {
    super(m, r, p, v, a);
    setOrientation(o);
    setTangentFriction(Friction.ZERO);
    setNormalFriction(Friction.ZERO);
    }


  /** 
   *  Returns the orientation of this body. The orientation is 
   *	a unit vector (i.e., <code>getOrientation().length() == 1.0</code>).
   */
  public Vector2 getOrientation()
    {
    return orientation;
    }

  /** Sets the orientation to <code>o</code>. The vector <code>o</code>
   *  needs not to be a unit vector. However, the resulting orientation
   *  vector is a unit vector that has the same direction as <code>o</code>.
   */
  public void setOrientation(Vector2 o)
    {
    orientation = o.unitVector();
    }

	/** Returns the tangent friction. */
	public Friction getTangentFriction()
		{
		return tangentFriction;
		}

	/** Sets the tangent friction to <code>f</code> */
	public void setTangentFriction(Friction f)
		{
		tangentFriction = (Friction)f.clone();
    tangentFrictionIsKinetic = false;
		}

	/** Returns the normal friction. */
	public Friction getNormalFriction()
		{
		return normalFriction;
		}

	/** Sets the normal friction to <code>f</code> */
	public void setNormalFriction(Friction f)
		{
		normalFriction = (Friction)f.clone();
    normalFrictionIsKinetic = false;
		}

		 
	/**
	 *	Returns the velocity of the body after <time>dt</time> seconds.
	 *  If the body is not frictionless, tangent and normal frictions 
	 *  slow down the velocity.
	 */	 
	public Vector2 futureVelocity(double dt)
		{
		Vector2 normal = (Vector2)orientation.clone();
		normal.rotate90();
	
		Vector2 tangentAcceleration = orientation.project(getAcceleration());
		Vector2 normalAcceleration = normal.project(getAcceleration());
				
		Vector2 tangentVelocity = orientation.project(getVelocity());
		Vector2 normalVelocity = normal.project(getVelocity());

    double g = (getOwner() != null ? 
    						getOwner().getAccelerationOfGravity() : 0.0);

		Vector2 cotangent = (Vector2)orientation.clone();
		cotangent.rotate180();		
		Vector2 conormal = (Vector2)orientation.clone();
		conormal.rotate270();		

		Vector2 movement = new Vector2(getPosition(), futurePosition(dt));
		Vector2 tangentMovement = orientation.project(movement);
		Vector2 normalMovement = normal.project(movement);

		tangentFrictionIsKinetic =
			frictionEpsilon.compare(tangentMovement.length(), 0.0) != 0;
		normalFrictionIsKinetic =
			frictionEpsilon.compare(normalMovement.length(), 0.0) != 0;

		if (!tangentFrictionIsKinetic  && 
		 		(tangentFriction.getStaticCoefficient() * g) 
		 		 > tangentAcceleration.length()) 
			{
			tangentAcceleration = (Vector2)Vector2.ZERO.clone();
			tangentVelocity = (Vector2)Vector2.ZERO.clone();
			}
		else 
			{
			cotangent.multiply(tangentFriction.getKineticCoefficient() * g);
			if (cotangent.isInTheSameQuadrant(tangentVelocity))
				cotangent.rotate180();
			tangentAcceleration.add(cotangent);
			}
		if (!normalFrictionIsKinetic  &&
			(normalFriction.getStaticCoefficient() * g) > normalAcceleration.length()) 
			{
			normalAcceleration = (Vector2)Vector2.ZERO.clone();
			normalVelocity = (Vector2)Vector2.ZERO.clone();
			}
		else 
			{
			conormal.multiply(normalFriction.getKineticCoefficient() * g);
			if (conormal.isInTheSameQuadrant(normalVelocity))
				conormal.rotate180();
			normalAcceleration.add(conormal);
			}
    tangentAcceleration.add(normalAcceleration);
    tangentVelocity.add(normalVelocity);

    double vx = tangentVelocity.getDeltaX();
    double vy = tangentVelocity.getDeltaY();
    double ax = tangentAcceleration.getDeltaX();
    double ay = tangentAcceleration.getDeltaY();
    vx = (ax * dt) + vx;
    vy = (ay * dt) + vy;
    return new Vector2(vx, vy);
		}		 
		 		 
  
  /** Returns a string that contains the values of the attributes. */
  public String toString()
    {
    return "orientation = " + orientation + " tangentFriction = " 
      + tangentFriction + " normalFriction = " + normalFriction
    	+ " " + super.toString();
    }
  
  }
  