/* 
 * Particle.java 1.1 2002-06-11
 *
 *  2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

import java.util.Collection;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 *  A fundamental particle. The particle can be owned by
 *  an instance of the class <code>Universe</code>. The particle's
 *  elasticity is used in collisions. The particle can be either corporeal 
 *  or ethereal, which determines whether it can collide or not (i.e., only 
 *  the corporeal particles can collide). However, the particle
 *  can report on collisions (whether they are corporeal or not) through
 *  the owning <code>Universe</code> object.
 * 
 *  @see Universe
 *  @since 1.1
 *  @version 1.1, 020611
 *  @author  Jouni Smed
 */

public abstract class Particle 
	{
  /** The universe that owns the particle. */
  private Universe owner;
  /** Is the particle corporeal or ethereal. */
  private boolean corporeal;
  /** Are the collisions notified. */
  private Collection collisionNotifications;


  /** 
   *  Constructs a particle that is corporeal, does not report on
   *  collisions and has no owner.  
   */
  protected Particle()
    {
    setOwner(null);
    setCorporeal();
    collisionNotifications = new LinkedList();
    }

  /** 
   *  Returns the owner of the particle or <code>null</code> if
   *  it has no owner.
   */
  public Universe getOwner()
    {
    return owner;
    }

  /** Sets the universe that owns the particle to <code>u</code>. */
  public void setOwner(Universe u)
    {
    owner = u;
    }
	
	
  /** 
   *  Returns <code>true</code> if the particle is corporeal or 
   *  <code>false</code> if the particle is ethereal.
   */
  public boolean isCorporeal()
    {
    return corporeal;
    }

  /** Sets the particle to be corporeal. */
  public void setCorporeal()
    {
    corporeal = true;
    }

  /** Sets the particle to be ethereal. */
  public void setEthereal()
    {
    corporeal = false;
    }


	/** 
	 *	Returns whether the collisions of and to this particle
	 *  by instances of the class <code>c</code> are notified.
	 *  Although only corporeal particles can have an actual 
	 *  impact in a collision, ethereal particles can report on
	 *  collisions.
	 *
	 *  @see #addCollisionNotification(Class)
	 *	@see Universe.CollisionNotification
	 */
	public boolean isCollisionNotified(Class c)
		{
		if (collisionNotifications.isEmpty()) return false;
		Iterator it = collisionNotifications.iterator();
		while (it.hasNext())
			{
			Class cc = (Class)it.next();
			if (cc.isAssignableFrom(c)) return true;
			}
		return false;
		}


	/** 
	 *	Returns whether the collisions of and to this particle
	 *  by the particle <code>p</code> are notified.
	 *  Although only corporeal particles can have an actual impact 
	 *  in a collision, ethereal particles can report on collisions.
	 */
	public boolean isCollisionNotified(Particle p)
		{
		if (collisionNotifications.isEmpty()) return false;
		Iterator it = collisionNotifications.iterator();
		while (it.hasNext())
			{
			Class cc = (Class)it.next();
			if (cc.isInstance(p)) return true;
			}
		return false;
		}


	/** 
	 *  Adds the class <code>c</code> to the collision notifications.
	 */
	public void addCollisionNotification(Class c)
		{
		collisionNotifications.add(c);
		}


	/** 
	 * 	Removes the class <code>c</code> from the collision notifications.
	 */
	public void removeCollisionNotification(Class c)
		{		
		boolean success = collisionNotifications.remove(c);
		}


	/** Removes all collision notifications. */
	public void clearCollisionNotifications()
		{
		collisionNotifications.clear();
		}


  /**
   *  Realizes an elastic collision for the particle and
   *  the body <code>other</code>. The linear momentum and kinetic 
   *  energy are conserved in the collision.
   *  The collision is realized regardless whether the body overlaps or
   *  crosses the particle or not. 
   *  @see #inelasticCollision(Body)
   */
  public abstract void elasticCollision(Body other);


  /**
   *  Realizes an inelastic collision for the particle and
   *  the body <code>other</code>. The linear momentum and kinetic 
   *  energy are lost in the collision. After the collision, the
   *  particle and the body are at rest (i.e., the velocity and acceleration 
   *  are zero).
   *  The collision is realized regardless whether the body overlaps or
   *  crosses the particle or not. 
   *  @see #elasticCollision(Body)
   */
  public abstract void inelasticCollision(Body other);



  /** Returns a string that contains the values of the attributes. */
  public String toString()
    {
    return "corporeal = " + isCorporeal();
    }

	}
