/* 
 * Point2.java 1.1 2002-06-04
 *
 *  2001-2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

import java.util.Comparator;

/**
 *  A two-element point of double-precision floating point x, y coordinates.
 *
 *  @version 1.1, 020604
 *  @author  Jouni Smed
 */
 
public class Point2 extends Tuple2
  {
  /** The origin residing in the coordinates (0.0, 0.0). */
  public static final Point2 ORIGIN = new Point2(0.0, 0.0);

  /**
   *  A comparator for the x coordinates of <code>Point2</code> objects. 
   *  The <code>compare</code> method returns a negative integer, zero, 
   *  or a positive integer as the x coordinate of the first argument is 
   *  less than, equal to, or greater than the x coordinate of the second.   
   */
  public static final Comparator HORIZONTAL = new Horizontal();
  
  /**
   *  A comparator for the y coordinates of <code>Point2</code> objects. 
   *  The <code>compare</code> method returns a negative integer, zero, 
   *  or a positive integer as the y coordinate of the first argument is 
   *  less than, equal to, or greater than the y coordinate of the second.   
   */
  public static final Comparator VERTICAL = new Vertical();
  
  /** Creates a new two-element point with the given coordinates. */
  public Point2(double x, double y)
    {
    super(x, y);
    }

  /** Returns the x coordinate. */
  public double getX() 
    {
    return getFirstElement();
    }

  /** Returns the y coordinate. */
  public double getY() 
    {
    return getSecondElement();
    }

  /** Sets the x coordinate to <code>x</code>. */
  public void setX(double x)
    {
    setFirstElement(x);
    }

  /** Sets the y coordinate to <code>y</code>. */
  public void setY(double y)
    {
    setSecondElement(y);
    }

  /**
   *  Sets the coordinates.
   *
   *  @param x The value of the x coordinate.
   *  @param y The value of the y coordinate.
   *  @see #set(Point2)
   **/
  public void set(double x, double y)
    {
    super.set(x, y);
    }

  /** Sets the coordinates to the same values as in <code>other</code>. */
  public void set(Point2 other)
    {
    super.set(other);
    }

  /**
   *  Adds values to the coordinates.
   *
   *  @param x The value added to the x coordinate.
   *  @param y The value added to the y coordinate.
   *  @see #add(Point2)
   */
  public void add(double x, double y)
    {
    super.add(x, y);
    }

  /** Adds the coordinates of <code>other</code> to the coordinates. */
  protected void add(Point2 other)
    {
    super.add(other);
    }

  /**
   *  Transfers the coordinates according to the vector <code>v</code>.
   *  After the transfer, <code>getX() = this.getX() + v.getDeltaX()</code> and
   *  <code>getY() = this.getY() + v.getDeltaY()</code>.
   */
  public void transfer(Vector2 v)
    {
    add(v);
    }
  
  /**
   *  Creates and returns a vector that corresponds the transfer
   *  from this point to the point <code>target</code>. In the new
   *  vector, <code>getDeltaX() = target.getX() - this.getX()</code> and
   *  <code>getDeltaY() = target.getY() - this.getY()</code>.
   *
   *  @see #vectorFrom(Point2)
   *  @see Vector2#Vector2(Point2,Point2)
   */
  public Vector2 vectorTo(Point2 target)
    {
    return new Vector2(this, target);
    }
  
  /**
   *  Creates and returns a vector that corresponds the transfer
   *  from the point <code>source</code> to this point. In the new
   *  vector, <code>getDeltaX() = this.getX() - source.getX()</code> and
   *  <code>getDeltaY() = this.getY() - source.getY()</code>.
   *
   *  @see #vectorTo(Point2)
   *  @see Vector2#Vector2(Point2,Point2)
   */
  public Vector2 vectorFrom(Point2 source)
    {
    return new Vector2(source, this);
    }

  /**
   *  A comparator for the x coordinates of <code>Point2</code> objects.
   *  @see Point2.Vertical
   */
  private static class Horizontal implements Comparator
    {
    /**
     *  Returns a negative integer, zero, or a positive integer as the
     *  x coordinate of the first argument is less than, equal to, or 
     *  greater than the x coordinate of the second.
     *  @throws ClassCastException If the arguments are not of the type
     *                             <code>Point2</code>.
     */
    public int compare(Object o1, Object o2)
      {
      double x1 = ((Point2)o1).getX();
      double x2 = ((Point2)o2).getX();
      return (x1 == x2 ? 0 : (x1 < x2 ? -1 : 1));
      }
    }

  /**
   *  A comparator for the y coordinates of <code>Point2</code> objects.
   *  @see Point2.Horizontal
   */
  private static class Vertical implements Comparator
    {
    /**
     *  Returns a negative integer, zero, or a positive integer as the
     *  y coordinate of the first argument is less than, equal to, or 
     *  greater than the y coordinate of the second.
     *  @throws ClassCastException If the arguments are not of the type
     *                             <code>Point2</code>.
     */
    public int compare(Object o1, Object o2)
      {
      double y1 = ((Point2)o1).getY();
      double y2 = ((Point2)o2).getY();
      return (y1 == y2 ? 0 : (y1 < y2 ? -1 : 1));
      }
    }
  }
