/* 
 * Tuple2.java 1.1 2002-05-06
 *
 *  2001-2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

/**
 *  A two-element tuple of double-precision floating point values.
 *
 *  @version 1.1, 020506
 *  @author  Jouni Smed
 */

public abstract class Tuple2 implements Cloneable
  {
  /** The first element of the tuple. */
  private double element1;
  /** The second element of the tuple. */
  private double element2;

  /** Creates a new two-element tuple with the given values.  */
  protected Tuple2(double v1, double v2)
    {
    element1 = v1;
    element2 = v2;
    }

  /** Returns the first element. */
  protected double getFirstElement()
    {
    return element1;
    }

  /** Returns the second element. */
  protected double getSecondElement()
    {
    return element2;
    }

  /** Sets the first element to the value <code>v</code>. */
  protected void setFirstElement(double v)
    {
    element1 = v;
    }

  /** Sets the second element to the value <code>v</code>. */
  protected void setSecondElement(double v)
    {
    element2 = v;
    }

  /**
   *  Sets the elements.
   *
   *  @param v1 The value of the first element.
   *  @param v2 The value of the second element.
   *  @see #set(Tuple2)
   **/
  protected void set(double v1, double v2)
    {
    element1 = v1;
    element2 = v2;
    }
    
  /** Sets the elements to the same values as in <code>other</code>. */
  protected void set(Tuple2 other)
    {
    element1 = other.getFirstElement();
    element2 = other.getSecondElement();
    }    

  /**
   *  Adds values to the elements.
   *
   *  @param v1 The value added to the first element.
   *  @param v2 The value added to the second element.
   *  @see #add(Tuple2)
   */
  protected void add(double v1, double v2)
    {
    element1 += v1;
    element2 += v2;
    }

  /** Adds the values of <code>other</code> to the elements. */
  protected void add(Tuple2 other)
    {
    element1 += other.getFirstElement();
    element2 += other.getSecondElement();
    }

  /** Negates the elements. */
  protected void negate()
    {
    negateFirstElement();
    negateSecondElement();
    }
    
  /** Negates the first element. */
  protected void negateFirstElement()
    {
    element1 = -element1;
    }

  /** Negates the second element. */
  protected void negateSecondElement()
    {
    element2 = -element2;
    }

  /** 
   *  Multiplies both elements with <code>value</code>. 
   *  @since 1.1
   */
  protected void multiply(double value)
    {
    element1 *= value;
    element2 *= value;
    }


  /** Swaps the elements. */
  protected void swap()
    {
    double temp = element2;
    element2 = element1;
    element1 = temp;
    }

  /** Returns the minimum of the elements. */
  protected double min()
    {
    return Math.min(element1, element2);
    }

  /** Returns the maximum of the elements. */
  protected double max()
    {
    return Math.max(element1, element2);
    }

  /** 
   *  Returns the sign of the first element. 
   *  @return A negative integer if the sign is minus, a positive integer
   *          if the sign is plus, or zero if the element is zero.
   *	@since 1.1
   */
  protected int firstElementSign()
    {
    return (element1 == 0.0 ? 0 : (element1 < 0.0 ? -1 : 1));
    }

  /** 
   *  Returns the sign of the second element. 
   *  @return A negative integer if the sign is minus, a positive integer
   *          if the sign is plus, or zero if the element is zero.
   *	@since 1.1
   */
  protected int secondElementSign()
    {
    return (element2 == 0.0 ? 0 : (element2 < 0.0 ? -1 : 1));
    }

  /** 
   *  Returns true if <code>o</code> is of the type <code>Tuple2</code> and 
   *  the elements of <code>o</code> are equal to the
   *  corresponding elements in this tuple.
   */
  public boolean equals(Object o)
    {
    if (!(o instanceof Tuple2)) return false;
    Tuple2 other = (Tuple2)o;
    return (this.element1 == other.getFirstElement() 
          & this.element2 == other.getSecondElement());
    }

  /** Returns a hash code value based on the elements of this tuple. */
  public int hashCode()
    {
    return (new Double(element1).hashCode() ^ new Double(element2).hashCode());
    }

  /** Creates and returns a copy of this tuple. */
  public Object clone()
    {
    try
      {
      return super.clone();
      }
    catch (CloneNotSupportedException e)
      {
      throw new RuntimeException("Cloning failed: " + e);
      }
    }

  /** Returns a string that contains the values of the elements. */
  public String toString()
    {
    return "(" + element1 + ", " + element2 + ")";
    }
  
  }
