/* 
 * Vector2.java 1.1 2002-09-01
 *
 *  2001-2002 Jouni Smed
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of the University of Turku nor the names of its
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package fi.utu.cs.physics;

/**
 *  A two-component vector of double-precision floating point values.
 *
 *  @version 1.1, 020901
 *  @author  Jouni Smed
 */

public class Vector2 extends Tuple2 implements Comparable
  {
  /** A vector with the length 0.0. */
  public static final Vector2 ZERO = new Vector2(0.0, 0.0);
    
  /** Creates a new two-component vector with the given values. */
  public Vector2(double dx, double dy)
    {
    super(dx, dy);
    }

  /** Creates a new two-component vector from the given points. In the 
   *  new vector, <code>this.getDeltaX() == x2 - x1</code>
   *  and <code>this.getDeltaY() == y2 - y1</code>.
   */
  public Vector2(double x1, double y1, double x2, double y2)
    {
    super(x2 - x1, y2 - y1);
    }

  /** 
   *  Creates a new two-component vector from the given points. In the 
   *  new vector, <code>this.getDeltaX() == target.getX() - source.getX()</code>
   *  and <code>this.getDeltaY() == target.getY() - source.getY()</code>.
   */ 
  public Vector2(Point2 source, Point2 target)
    {
    super(target.getX() - source.getX(), target.getY() - source.getY());
    }

  /** Returns the horizontal component. */
  public double getDeltaX()
    {
    return getFirstElement();
    }

  /** Returns the vertical component. */
  public double getDeltaY() 
    {
    return getSecondElement();
    }

  /** Sets the horizontal component to the value <code>dx</code>. */
  public void setDeltaX(double dx)
    {
    setFirstElement(dx);
    }

  /** Sets the vertical component to the value <code>dy</code>. */
  public void setDeltaY(double dy)
    {
    setSecondElement(dy);
    }

  /**
   *  Sets the horizontal and vertical component.
   *
   *  @param dx The value of the horizontal component.
   *  @param dy The value of the vertical component.
   *  @see #set(Point2,Point2)
   *  @see #set(Vector2)
   **/
  public void set(double dx, double dy)
    {
    super.set(dx, dy);
    }

  /**
   *  Sets the horizontal and vertical components to correspond
   *  the transfer from the point <code>source</code> to the point 
   *  <code>target</code>. After the operation,
   *  <code>this.getDeltaX() = target.getX() - this.getX()</code> and
   *  <code>this.getDeltaY() = target.getY() - this.getY()</code>.
   *
   *  @see #set(Vector2)
   **/
  public void set(Point2 source, Point2 target)
    {
    set(target.getX() - source.getX(), target.getY() - source.getY());
    }

  /** 
   *  Sets the horizontal and vertical components to the same 
   *  values as in <code>other</code>. 
   */
  public void set(Vector2 v)
    {
    super.set(v);
    }
  
  /** Returns the length of this vector. */
  public double length()
    {
    return Math.sqrt(lengthSquared());
    }

  /** Returns the squared length of this vector. */
  public double lengthSquared()
    {
    return ((getDeltaX() * getDeltaX()) + (getDeltaY() * getDeltaY()));
    }

  /**
   *  Adds values to the components.
   *
   *  @param dx The value added to the horizontal component.
   *  @param dy The value added to the vertical component.
   *  @see #add(Vector2)
   */
  public void add(double dx, double dy)
    {
    super.add(dx, dy);
    }

  /** Adds the components of <code>other</code> to the components. */
  public void add(Vector2 other)
    {
    super.add(other);
    }


	/** 
	 *	Multiplies the vector with the scalar <code>s</code>.
	 *  @since 1.1
	 */
	public void multiply(double s)
		{
		super.multiply(s);
		}
	

	/** 
	 *	Returns the scalar product of this vector and 
	 *  the vector <code>other</code>.
	 *  @since 1.1
	 */
	public double scalarProduct(Vector2 other)
		{
		return (this.getDeltaX() * other.getDeltaX()) 
				 + (this.getDeltaY() * other.getDeltaY());
		}


	/** 
	 *	Returns a unit vector corresponding to this vector.
	 *  More formally, <code>this.unitVector().length == 1.0</code>
	 *  and <code>this.angle(this.unitVector()) == 0.0</code>.
	 *  @since 1.1
	 */
	public Vector2 unitVector()
		{
		Vector2 unitVector = (Vector2)this.clone();
		unitVector.multiply(1.0 / this.length());
		return unitVector;
		}

	/** 
	 *	Returns a projection of the vector <code>other</code> to this
	 *  vector.
	 *  @since 1.1
	 */
	public Vector2 project(Vector2 other)
		{
		Vector2 projection = (Vector2)this.clone();
		projection.multiply(this.scalarProduct(other) / this.scalarProduct(this));
		return projection;
		}

	/** 
	 *	Returns the angle between this vector and the vector
	 *  <code>other</other>. 
	 *  @return angle in radians from the range 
	 *								[<code>0.0</code>, <code>Math.PI</code>]
	 *  @see #rotate(double)
	 *	@since 1.1
	 */
	public double angle(Vector2 other)
		{
		return Math.acos(scalarProduct(other) / 
						(this.length() * other.length()));
		}

  /**
   *  Creates and returns a point that is transfered 
   *  from the point <code>source</code> according to this vector.
   *  In the new point, <code>getX() = source.getX() + this.getDeltaX()</code>
   *  and <code>getY() = source.getY() + this.getDeltaY()</code>.
   */
  public Point2 transfer(Point2 source)
    {
    return new Point2(source.getX() + getDeltaX(), source.getY() + getDeltaY());
    }
  
  /** 
   *  Rotates the vector 90 degrees counterclockwise. 
   *  @see #rotate(double)
   *  @see #rotate180()
   *  @see #rotate270()
   */
  public void rotate90()
    {
    swap();
    negateSecondElement();
    }
  
  /** 
   *  Rotates the vector 180 degrees counterclockwise. 
   *  @see #rotate(double)
   *  @see #rotate90()
   *  @see #rotate270()
   */
  public void rotate180()
    {
    negate();
    }

  /** 
   *  Rotates the vector 270 degrees counterclockwise. 
   *  @see #rotate(double)
   *  @see #rotate90()
   *  @see #rotate180()
   */
  public void rotate270()
    {
    swap();
    negateFirstElement();
    }

  /** 
   *  Rotates the vector counterclockwise. 
   *  @param r The amount of rotation in radians.
	 *  @see #angle(Vector2)
   *  @see #rotate90()
   *  @see #rotate180()
   *  @see #rotate270()
   */
  public void rotate(double r)
    {
    double x = getDeltaX() * Math.cos(r) - getDeltaY() * Math.sin(r);
    double y = getDeltaX() * Math.sin(r) + getDeltaY() * Math.cos(r);
    set(x, y);
    }

	/**
	 *	Returns whether the vector <code>other</code> resides in the 
	 *  same coordinate quadrant as this vector. 
	 *	@since 1.1
	 */
	public boolean isInTheSameQuadrant(Vector2 other)
		{
		return ((this.firstElementSign() == other.firstElementSign()) &
				(this.secondElementSign() == other.secondElementSign()));
		}
	

  /**
   *  Returns a negative integer, zero, or a positive integer as the
   *  length of this vector is less than, equal to, or greater than 
   *  the length of vector <code>o</code>.     
   *  @throws ClassCastException If the argument is not of the type
   *                             <code>Vector2</code>.
   */
  public int compareTo(Object o)
    {
    Vector2 other = (Vector2)o;
    double thisLength = lengthSquared();
    double otherLength = other.lengthSquared();
    return (thisLength == otherLength ? 0 : 
           (thisLength < otherLength ? -1 : 1));
    }

  }
